# to do
import os
from copy import deepcopy
from typing import Callable, List, Tuple

from lxml import etree

from sastadev import compounds
from sastadev.conf import settings
from sastadev.lexicon import informlexicon
from sastadev.metadata import (filled_pause, fstoken, intj, janeenou, longrep,
                               repeated, repeatedjaneenou, repeatedseqtoken,
                               shortrep, substringrep, unknownsymbol,
                               unknownword)
from sastadev.phonetics import phoneticise
from sastadev.sastatoken import Token
from sastadev.sastatypes import Nort, SynTree
from sastadev.stringfunctions import deduplicate, string2list
from sastadev.tblex import asta_recognised_wordnode
from sastadev.treebankfunctions import (all_lower_consonantsnode, find1,
                                        getattval, getnodeyield,
                                        lastmainclauseof, openclasspts)

nodetype = etree._Element

positionatt = 'end'

xmetaxpath = './/xmeta'

samplesizemdvalues = {repeatedjaneenou,
                      shortrep, intj, unknownsymbol, filled_pause}
mlumdvalues = {repeated, repeatedseqtoken, longrep, unknownword, substringrep, janeenou,
               fstoken}


class DupInfo:
    '''
    The class *DupInfo* defines objects for storing information about duplicate and
    partially duplicate words, both for long duplications (>=50%) and for short
    duplications (<50%). In addition it contains a list of words in incomplete sentences
    '''

    def __init__(self, longdups=dict(), shortdups=dict(), icsws=[]):
        self.longdups = longdups
        self.shortdups = shortdups
        self.icsws = icsws  # nodes for words in incomplete sentences

    def __str__(self):
        result = str(self.longdups) + ';' + \
            str(self.shortdups) + str(self.icsws)
        return result

    def merge(self, dupinfo):
        '''
        The method *merge* merges two DupInfo objects by dictmerging their longdups
        dictionaries, their shortdups dictionaries, and by concatenating the lists of
        words form incomplete sentences.
        '''
        newdupinfo = deepcopy(self)
        newdupinfo.longdups = dictmerge(self.longdups, dupinfo.longdups)
        newdupinfo.shortdups = dictmerge(self.shortdups, dupinfo.shortdups)
        newdupinfo.icsws = self.icsws + dupinfo.icsws
        return newdupinfo

    def get_chaintail(self, i):
        if i in self.shortdups:
            result = self.get_chaintail(self.shortdups[i])
        elif i in self.longdups:
            result = self.get_chaintail(self.longdups[i])
        else:
            result = i
        return result


def dictmerge(dict1, dict2):
    '''
    The function *dictmerge* merges two dictionaries dict1 and dict2 into a new
    dictionary, but only for non-conflicting and non-identical key-value pairs.
    '''
    newdict = deepcopy(dict1)
    for el in dict2:
        if el in newdict:
            if newdict[el] != dict2[el]:
                settings.LOGGER.error(
                    'Conflicting values for {}: {}: {} not included'.format(
                        el, newdict[el], dict2[el]
                    )
                )
            else:
                settings.LOGGER.warning(
                    'Duplicate values for {}: {} = {}'.format(el, newdict[el], dict2[el]))
        else:
            newdict[el] = dict2[el]
    return newdict


normalisedict = {'c': 'k'}

#: The constant *unwantedtokenlist* contains symbols and symbol sequences that should
#: be ignored for the sample size
unwantedtokenlist = ['-', '--', '#', '–', '\u2013', '\u2014', '\u2015']

incomplete_zijn = '''
// node[node[ @rel = "hd" and @lemma="zijn"] and
        not (node[ @rel="predc" or @rel="vc" or @rel="ld" or ( @rel="mod" and @lemma="er") or
             node[ @rel = "mod" and node[ @rel = "hd" and (@lemma="voor")]]
           ])
         and count(. // node[ @ pt]) < 6
       ]
'''

incomplete_hebben = '''
//node[node[@rel="hd" and @lemma="hebben"] and
       (not(node[@rel="obj1" or @rel="vc" or @rel="svp"]) )
       and count(.//node[@pt]) < 6
      ]
'''

incomplete_eentw = '''
//node[node[(@rel!="det" and @rel!="hd" and @rel!="mwp" and @rel!="--") and @lemma="één"] and not(node[@lemma="er"])]
'''

#: The variable *incompletexpaths* contains Xpath queries to identify incomplete
#: phrases in an utterance. It currently contains 3 such queries:
#:
#:   * one for clauses headed by *hebben*
#:   * one for clauses headed by *zijn*
#:   * one for incomplete noun phrases just consisting of *een* (often wrongly interpreted
#:     by Alpino as a full NP consisting of the numeral *één*.)
#:
incompletexpaths = [incomplete_hebben, incomplete_zijn, incomplete_eentw]

space = ' '

janeenouset = set()
janeenouset = {'ja', 'nee', 'nou'}

janeeset = set()
janeeset = {'ja', 'nee'}


def getposition(nort):
    if nort is None:
        result = None
    elif isinstance(nort, Token):
        result = nort.pos
    elif isinstance(nort, nodetype):
        result = getattval(nort, positionatt)
    else:
        result = ('??')
    return result


def getword(nort):
    if nort is None:
        result = ''
    elif isinstance(nort, Token):
        result = nort.word
    elif isinstance(nort, nodetype):
        result = getattval(nort, 'word')
    else:
        result = '**'
    return result


def onvolledig(stree):
    results = []
    if incomplete(stree):
        results += [stree]
    return results


def incomplete(stree):
    result = False
    for query in incompletexpaths:
        if result:
            break
        else:
            allmatches = stree.xpath(query)
            matches = [m for m in allmatches if m == lastmainclauseof(stree)]
            result = matches != []
    return result


def incompletetreeleaves(stree: SynTree) -> List[SynTree]:
    '''
    The function *incompletetreeleaves returns a list of all nodes for words that are
    part of an incomplete sentence. A sentence is incomplete if it matches a query from the list of queries in the variable *incompletexpaths*.

    .. autodata:: dedup::incompletexpaths

    '''
    results = []
    for query in incompletexpaths:
        allmatches = stree.xpath("." + query)
        matches = [m for m in allmatches if m != lastmainclauseof(stree)]
        for m in matches:
            results += getnodeyield(m)
    return results


def findcorrections(nodelist):
    wordlist = [getattval(n, 'word') for n in nodelist]
    resultlist = []
    lnodelist = len(nodelist)
    for i in range(lnodelist - 1):
        n1 = nodelist[i]
        n2 = nodelist[i + 1]
        if getattval(n1, 'pt') == getattval(n2, 'pt'):
            resultlist.append((n1, n2))
    return resultlist


def isxxx(node):
    theword = getword(node)
    result = theword.lower() in {'xxx', 'yyy', 'www'}
    return result


def isfilledpausenort(nort: Nort) -> bool:
    '''
    The function *isfilledpausenort* returns the result of the function *isfilledpause* applied to the *word* of *nort*.

      * .. autofunction:; dedup::isfilledpause
    '''

    theword = getword(nort)
    result = isfilledpause(theword)
    return result


def getfilledpauses(nortlist: List[Nort]) -> List[Nort]:
    '''
    The function *getfilledpauses returns Norts that are in nortlist for which the
    function *isfilledpausenort* yields True.

      * .. autofunction:: dedup::isfilledpausenort

    '''
    resultlist = [tok for tok in nortlist if isfilledpausenort(tok)]
    return resultlist


def infilledpauses(word: str) -> bool:
    return (word in filledpauseslexicon)


def isfilledpause(word: str) -> bool:
    '''
    The function *isfilledpause* determines whether *word* is a filled pause. This is
    the case if

    * the lower case variant of *word* is contained in the *filledpauseslexicon*
    * or else if a deduplicated version of the lower case variant of *word* is contained in the *filledpauseslexicon*

    Otherwise the function returns False.

    Deduplication is achived by the function *deduplicate* from the module *stringfunctions*

    '''
    lcword = word.lower()
    if infilledpauses(lcword):
        result = True
    else:
        swords = deduplicate(lcword, infilledpauses)
        if (swords != []):
            result = True
        else:
            result = False
    return result


def getfilledpausesposlist(nortlist):
    results = []
    for token in nortlist:
        theword = getword(token)
        if isfilledpause(theword):
            thepos = getposition(token)
            results.append(thepos)
    return results


def remove_duplicates(wl):
    stop = False
    result = wl
    lwl = len(wl)
    ml = lwl // 2
    tobedeleted = []
    for curlen in range(ml, 0, -1):
        for startpos in range(0, lwl - 2 * curlen + 1, 1):
            if isduplicate(wl[startpos:startpos + curlen],
                           wl[startpos + curlen:startpos + 2 * curlen]):
                newwl = wl[:startpos] + wl[startpos + curlen:]
                result = remove_duplicates(newwl)
                stop = True
                break
        if stop:
            break
    return result


def isnortsubstring(n1, n2):
    w1 = getword(n1)
    w2 = getword(n2)
    lcw1 = w1.lower()
    lcw2 = w2.lower()
    result = lcw1 in lcw2 and len(w1) / len(w2) > 0.5
    return result


def find_substringduplicates2(wl: List[Nort]) -> Tuple[List[Nort], DupInfo]:
    '''
    The function *find_substringduplicates2* finds nodes with words that are a
    substring of the word  of the successor node, and it creates a DupInfo object that
    contains a dictionary  with <position substring word, position successor word> items.

    '''
    dupmapping = dict()
    result = []
    lwl = len(wl)
    for i in range(lwl - 1):
        curtoken = wl[i]
        nexttoken = wl[i + 1]
        curlcword = getword(curtoken).lower()
        if isnortsubstring(curtoken, nexttoken) and not (informlexicon(curlcword)):
            duppos = getposition(curtoken)
            origpos = getposition(nexttoken)
            dupmapping[duppos] = origpos
            result.append(curtoken)
    alldupinfo = DupInfo(dupmapping, dict())
    return result, alldupinfo


def find_simpleduplicates(wl):
    result, _ = find_simpleduplicates2(wl)
    return result


def find_simpleduplicates2(wl: List[Nort]) -> Tuple[List[Nort], DupInfo]:
    '''
    The function *find_simpleduplicates2* identifies each Nort that is a duplicate of
    its successor. It returns a list of these Norts and a dictionary of
    <position of the duplicate: position of its successor> items as the longdups part in
    a DupInfo object.

    '''
    dupmapping = dict()
    result = []
    lwl = len(wl)
    for i in range(lwl - 1):
        if isnortduplicate([wl[i]], [wl[i + 1]]):
            duppos = getposition(wl[i])
            origpos = getposition(wl[i + 1])
            dupmapping[duppos] = origpos
            result.append(wl[i])
    alldupinfo = DupInfo(dupmapping, dict())
    return result, alldupinfo


def getreptokenpos(nort, dupinfo):
    nortpos = getposition(nort)
    if nortpos in dupinfo:
        result = str(dupinfo[nortpos])
    else:
        result = ''
    return result


def find_duplicates(wl):
    result, _ = find_duplicates2(wl)
    return result


# applies to a sequence of Lassy word nodes or token  nodes (Nort)
def find_duplicates2(wl: List[Nort]) -> Tuple[List[Nort], DupInfo]:
    '''
    The function *find_duplicates2* identifies each Nort sequence  that is a duplicate of
    its successor Nort sequence. It returns a list of these Norts and a dictionary of  the
    position of the duplicate and the positions of its successor items as the longdups
    part in a DupInfo  object.

    '''

    dupmapping = dict()
    alldupinfo = DupInfo()
    stop = False
    result = wl
    lwl = len(wl)
    ml = lwl // 2
    result = []
    for curlen in range(ml, 1, -1):  # minimum length is 2
        # for startpos in range(0, lwl-2*curlen+1,1):
        for startpos in range(lwl - 2 * curlen, -1,
                              -1):  # find dup seqeunces starting at the rightmost position
            if isnortduplicate(wl[startpos:startpos + curlen],
                               wl[startpos + curlen:startpos + 2 * curlen]):
                result = [wl[p] for p in range(startpos, startpos + curlen)]
                for p in range(startpos, startpos + curlen):
                    duppos = getposition(wl[p])
                    origpos = getposition(wl[p + curlen])
                    dupmapping[duppos] = origpos
                    alldupinfo = DupInfo(dupmapping, dict())
                newwl = wl[:startpos] + wl[startpos + curlen:]
                restresult, restdupinfo = find_duplicates2(newwl)
                result += restresult
                alldupinfo = alldupinfo.merge(restdupinfo)

                stop = True
                break
        if stop:
            break
    return result, alldupinfo


def find_janeenouduplicates(wl):
    result, _ = find_janeenouduplicates2(wl)
    return result


def find_janeenouduplicates2(wl):
    resultlist = []
    dupmapping = dict()
    lwl = len(wl)
    for i in range(lwl - 1):
        wlip = getposition(wl[i])
        wli1p = getposition(wl[i + 1])
        wliw = getword(wl[i])
        wli1w = getword(wl[i + 1])
        lcwliw = wliw.lower()
        lcwli1w = wli1w.lower()
        if lcwliw in janeenouset and lcwliw == lcwli1w:
            resultlist.append(wl[i])
            dupmapping[wlip] = wli1p
    dupinfo = DupInfo(longdups=dupmapping, shortdups=dict())
    return resultlist, dupinfo


def normalisestring(str1: str) -> str:
    '''
    The function *normalisestring* carries out normalisation by means of the function
    *phoneticise* from the *phonetics* module:

    .. autofunction:: phonetics::phoneticise

    '''
    result = phoneticise(str1)
    return result


def isnortduplicate(tlist1: List[Nort], tlist2: List[Nort]) -> bool:
    '''
    The function *isnortduplicate* determines whether tlist1 is a duplicate of tlist2,
    which is the case if the lists have equal length and each normalised word of
    tlist1 is equal to or a prefix of the corresponding normalised word in tlist2.

    Normalisation is carried out to be robust against certain spelling variations and
    is taken care of by the function *normalisestring*:

    *  .. autofunction:: dedup::normalisestring

    '''
    result = True
    ltlist1 = len(tlist1)
    ltlist2 = len(tlist2)
    result = ltlist1 == ltlist2
    if result:
        for i in range(ltlist1):
            lcword1 = getword(tlist1[i]).lower()
            lcword2 = getword(tlist2[i]).lower()
            nlcword1 = normalisestring(lcword1)
            nlcword2 = normalisestring(lcword2)
            result = result and ((nlcword1 == nlcword2)
                                 or nlcword2.startswith(nlcword1))
    return result


def nextnode(node, nodes):
    for i, n in enumerate(nodes):
        if n == node:
            if i + 1 < len(nodes):
                return nodes[i + 1]
            else:
                return None


def nodesfindjaneenou(nodes):
    janees = [n for n in nodes if getattval(n, 'lemma') in {'ja', 'nee'}]
    nous = [n for n in nodes if getattval(n, 'lemma') == 'nou' and (
        getattval(n, 'rel') in {'mwp', 'tag', 'cnj'}
        or getattval(nextnode(n, nodes), 'lemma') in {'ja', 'nee'})]
    results = janees + nous
    return results


def treefindjaneenou(stree):
    janees = stree.xpath('.//node[@lemma="ja" or @lemma="nee"]')
    nous = stree.xpath(
        './/node[@lemma="nou" and (@rel="mwp" or @rel="tag" or @rel="cnj" )] ')
    results = janees + nous
    return results


def findjaneenou(nortlist):
    resultlist = findnodefromset(nortlist, janeenouset)
    return resultlist


def findjanee(nortlist):
    resultlist = findnodefromset(nortlist, janeeset)
    return resultlist


def findnodefromset(nortlist, wordset):
    resultlist = []
    for node in nortlist:
        theword = getword(node)
        lctheword = theword.lower()
        if lctheword in wordset:
            resultlist.append(node)
    return resultlist


def isduplicate(wlist1, wlist2):
    result = wlist1 == wlist2
    return result


def cleantokenlist(tokenlist, tobedeleted):
    cleanlist = [t for t in tokenlist if getposition(t) not in tobedeleted]
    cleanstrlist = [getword(t) for t in cleanlist]
    cleanstr = space.join(cleanstrlist)
    return cleanstr


def correct(stree):
    correct1xpath = './/node[@cat="top" and node[(@cat="smain" or @cat="sv1" or @cat="whq" or @cat="whsub")] and  count(node[@cat])=1]'
    correct2xpath = './/node[@cat="top" and node[@cat="du" and node[@rel="dlink" or @rel="tag"] and node[(@cat="smain" or @cat="sv1" or @cat="whq" or @cat="whsub") and @rel="nucl"] ]]'
    correct3xpath = './/node[@cat="top" and node[@cat="du" and node[@cat="conj" and count(node[(@cat="smain" or @cat="sv1" or @cat="whq"  or @cat="whsub")])>1] ]]'
    correct4xpath = './/node[@cat="top" and  node[@cat="conj" and count(node[(@cat="smain" or @cat="sv1" or @cat="whq"  or @cat="whsub")])>1] ]'
    correct5xpath = './/node[@cat="top" and  node[@cat="du" and count(node)=2 and node[@rel="dp" or @rel="tag" and @end<../node[@rel="nucl"]/@end] and node[(@cat="smain" or @cat="sv1" or @cat="whq")]] ]'
    matches1 = stree.xpath(correct1xpath)
    matches2 = stree.xpath(correct2xpath)
    matches3 = stree.xpath(correct3xpath)
    matches4 = stree.xpath(correct4xpath)
    matches5 = stree.xpath(correct5xpath)
    matches = matches1 + matches2 + matches3 + matches4 + matches5
    results = []
    for m in matches:
        if not incomplete(m):
            results.append(m)
    return results


def mlux(stree: SynTree) -> List[SynTree]:
    '''
    The function *mlux* determines which nodes for word in *stree* should be included
    for the computation of the mean length utterance (mlu) by applying the function
    *mlux2* to *stree*. The latter function returns a node list and metadata on the
    excluded word nodes.

    .. autofunction:: dedup::mlux2

    '''
    result, _ = mlux2(stree)
    return result


def mlux2(stree: SynTree) -> Tuple[List[SynTree], DupInfo]:
    '''
    The function *mlux2* returns a tuple consisting of a list of nodes for words that
    should be included in computing the mean length utterance (mlu), and information
    about duplications in a DupInfo object.

    * It first obtains a list of nodes for words, in the right surface order, of *stree*
    * It filters those nodes that are found by *samplesize* (:ref:`A045_X`)
    * The function maintains two variables:

      * **cleantokennodelist**: the list of nodes still in the utterance
      * **resultnodelist**: the list of nodes that are in the result of this function
      * Each node that is added to the resultnodelist is removed from the cleantokennodelist

    * It updates these variables for  all nodes  if any of the nodes has *word* equal to *xxx*,  *yyy* or  *www*
    * It updates these variables for nodes that have been found by an earlier correction process: these are listed in the metadata.
    * it updates these variables for nodes for  unknown words that are of an open class part of speech
    * it updates these variables for nodes for nodes for the words *ja*, *nee*, *nou*
    * it updates these variables for nodes for interjections (*pt=tsw*)
    * it updates these variables for nodes for filledpauses as found in the :ref:`filledpauseslexicon`
    * it updates these variables for nodes for words and word sequences that are duplicated, using the function *find_simpleduplicates2*:

       * .. autofunction:: dedup::find_simpleduplicates2

    * it updates these variables for nodes for word sequences that are duplicated, using the function *find_duplicates2*:

       * .. autofunction:: dedup::find_duplicates2

    * it updates these variables for nodes for words the prefix of which is a repetition of its successor, where the prefix is larger than 50% of the length of its successor (long duplications). It does so by means of the function *getprefixwords2*:

       * .. autofunction:: dedup::getprefixwords2

   * it updates these variables for nodes for unknown words that are a substring of their successor. It does so using the function  *find_substringduplicates2*:

       * .. autofunction:: dedup::find_substringduplicates2

   * it updates these variables for nodes for words that consist of consonants only
   * it updates these variables for nodes for words in incomplete sentences.
   Determining the incompleteness of a sentence is very difficult and is so far only done for a limited number of sentence types. It uses the function *incompletetreeleaves* for this purpose:

      * .. autofunction:: dedup::incompletetreeleaves

   The function does not (yet) deal with:
    *  false starts, e.g. word + *nee* / *eh* word;  w of pos1 w of pos1
    * the sequence *of nee*
    * *dus* als stopwoordje

    '''
    debug = False
    if debug:
        etree.dump(stree)
    resultnodelist = []
    alldupinfo = DupInfo()
    tokennodelist = getnodeyield(stree)
    excludednodes, dupinfo = samplesize2(stree)
    cleantokennodelist = [n for n in tokennodelist if n not in excludednodes]
    alldupinfo = alldupinfo.merge(dupinfo)

    # remove all if xxx/yyy/www occurs ; this can be done here
    xxxfound = any([isxxx(n) for n in cleantokennodelist])
    if xxxfound:
        resultnodelist = cleantokennodelist
        cleantokennodelist = []

    # add results that have been found earlier by reduce in correction and that are now in the metadata
    # and exclude these nodes for further processing
    if cleantokennodelist != []:
        mdnodes = []
        mlumds = stree.xpath(xmetaxpath)
        for mlumd in mlumds:
            if 'value' in mlumd.attrib and mlumd.attrib['value'] in mlumdvalues:
                tokenbeginstr = mlumd.attrib['annotatedposlist']
                tokenbegins = string2list(tokenbeginstr)
                for tokenbegin in tokenbegins:
                    nodexpath = './/node[@pt and @begin="{}"]'.format(
                        tokenbegin)
                    newnode = find1(stree, nodexpath)
                    if newnode is not None:
                        mdnodes.append(newnode)
                    else:
                        settings.LOGGER.error(
                            'Metadata node not found in tree: md.begin={}'.format(
                                tokenbegin))
                        etree.dump(stree)
        excludednodes += mdnodes
        cleantokennodelist = [
            n for n in cleantokennodelist if n not in excludednodes]
        resultnodelist += mdnodes

    # remove unknown words if open class
    unknown_words = [n for n in cleantokennodelist if getattval(n, 'pt') in openclasspts
                     and not (asta_recognised_wordnode(n))]
    resultnodelist += unknown_words
    cleantokennodelist = [
        n for n in cleantokennodelist if n not in unknown_words]

    # ASTA sec 6.3 p. 11
    # remove ja nee nou
    # janeenoulist = findjaneenou(cleantokennodelist)

    janeenoulist = nodesfindjaneenou(cleantokennodelist)
    resultnodelist += janeenoulist
    cleantokennodelist = [
        n for n in cleantokennodelist if n not in janeenoulist]

    # remove false starts maybe word + nee / of nee / eh word; of w of pos1 w of pos1
    # remove of nee

    # remove tsw incl goh och hé oke
    tswnodes = [n for n in cleantokennodelist if getattval(n, 'pt') == 'tsw']
    resultnodelist += tswnodes
    cleantokennodelist = [n for n in cleantokennodelist if n not in tswnodes]

    # remove other filled pauses
    fpnodes = [n for n in cleantokennodelist if
               getattval(n, 'lemma') in filledpauseslexicon]
    resultnodelist += fpnodes
    cleantokennodelist = [n for n in cleantokennodelist if n not in fpnodes]

    # simple duplicates
    dupnodelist, dupinfo = find_simpleduplicates2(cleantokennodelist)
    resultnodelist += dupnodelist
    alldupinfo = alldupinfo.merge(dupinfo)
    cleantokennodelist = [
        n for n in cleantokennodelist if n not in dupnodelist]

    # for debugging
    # print(showtns(cleantokennodelist))
    dupnodelist, dupinfo = find_duplicates2(cleantokennodelist)
    resultnodelist += dupnodelist
    alldupinfo = alldupinfo.merge(dupinfo)
    cleantokennodelist = [
        n for n in cleantokennodelist if n not in dupnodelist]

    # find prefix herhalingen >= 50%
    def cond(x, y):
        return len(cleanwordofnort(x)) / len(cleanwordofnort(y)) > 0.5

    prefixnodes, dupinfo = getprefixwords2(cleantokennodelist, cond)
    resultnodelist += prefixnodes
    alldupinfo = alldupinfo.merge(dupinfo)
    cleantokennodelist = [
        n for n in cleantokennodelist if n not in prefixnodes]

    # find unknown words that are a substring of their successor
    substringnodes, dupinfo = find_substringduplicates2(cleantokennodelist)
    alldupinfo = alldupinfo.merge(dupinfo)
    cleantokennodelist = [
        n for n in cleantokennodelist if n not in prefixnodes]

    # corrections = findcorrections(cleantokennodelist)
    # if corrections != []:
    #    cleanwordlist = [getattval(n, 'word') for n in cleantokennodelist]
    #    print(space.join(cleanwordlist), file=testfile)
    # for (w, corr) in corrections:
    #    print('--', getattval(w, 'word'), getattval(corr, 'word'), file=testfile)

    # remove dus als stopwoordje

    # remove words that consist of consonants only
    resultnodelist = [n for n in resultnodelist if not (
        all_lower_consonantsnode(n))]

    # remove words in incomplete sentences
    isws = incompletetreeleaves(stree)
    pureisws = [n for n in isws if n in cleantokennodelist]
    resultnodelist += pureisws
    alldupinfo.icsws = pureisws
    cleantokenlist = [n for n in cleantokennodelist if n not in pureisws]

    return resultnodelist, alldupinfo


def getrepeatedtokens(tokenlist, repeatingtokens):
    '''

    :param tokenlist:
    :param repeatingtokens: list of tokens that are repeating a token in tokenlist
    :return: dictionary of repeatingtoken: repeatedtoken pair
    '''
    repeatedtokens = {}
    ltokenlist = len(tokenlist)
    for i in range(ltokenlist):
        if tokenlist[i] in repeatingtokens:
            repeatedtokens[tokenlist[i]] = tokenlist[i + 1]
    return repeatedtokens


def cleanwordofnort(token):
    word = getword(token)
    result = cleanwordof(word)
    return result


def cleanwordof(word):
    lcword = word.lower()
    return lcword


def isnortprefixof(node1, node2):
    w1 = getword(node1)
    cw1 = cleanwordof(w1)
    w2 = getword(node2)
    cw2 = cleanwordof(w2)
    result = isprefixof(cw1, cw2)
    return result


def isprefixof(cw1, cw2):
    ncw1 = normalisestring(cw1)
    ncw2 = normalisestring(cw2)
    result = ncw1 != ncw2 and ncw2.startswith(ncw1)
    return result


def getprefixwords(wlist, cond):
    result, _ = getprefixwords2(wlist, cond)
    return result


def getprefixwords2(wlist: List[Nort],
                    cond: Callable[[Nort, Nort], bool]) \
        -> Tuple[List[Nort], DupInfo]:
    '''
    The function *getprefixwords2* finds nodes with words that are a prefix of the word
    of the successor node, and it creates a DupInfo object that contains a dictionary
    with <position prefixword, position successor word> items.

    '''
    resultlist = []
    dupmapping = dict()
    lwlist = len(wlist) - 1
    tokenctr = lwlist
    while tokenctr > 0:
        repctr = tokenctr - 1
        while repctr >= 0 and isnortprefixof(wlist[repctr], wlist[tokenctr]):
            if cond(wlist[repctr], wlist[tokenctr]):
                resultlist.append(wlist[repctr])
                reppos = getposition(wlist[repctr])
                tokenpos = getposition(wlist[tokenctr])
                dupmapping[reppos] = tokenpos
            repctr -= 1
        tokenctr = repctr
    dupinfo = DupInfo(dict(), dupmapping)
    return resultlist, dupinfo


def newgetprefixwords2(wlist, cond):
    resultlist = []
    dupmapping = dict()
    lwlist = len(wlist) - 1
    tokenctr = lwlist
    while tokenctr > 0:
        repctr = tokenctr - 1
        while repctr >= 0:
            wr = wlist[repctr]
            wt = wlist[tokenctr]

            ok = isnortprefixof(wlist[repctr], wlist[tokenctr])
            # @@@hier @@
            if cond(wlist[repctr], wlist[tokenctr]):
                resultlist.append(wlist[repctr])
                reppos = getposition(wlist[repctr])
                tokenpos = getposition(wlist[tokenctr])
                dupmapping[reppos] = tokenpos
            repctr -= 1
        tokenctr = repctr
    dupinfo = DupInfo(dict(), dupmapping)
    return resultlist, dupinfo


def isnamenort(node):
    theword = getword(node)
    result = theword[0].lower() != theword[0]
    return result


compoundxpath = './/node[@his="compound"]'
wordxpath = './/node[@pt and @pt!="let"]'


def neologisme(stree):
    results = []
    thecompounds = stree.xpath(compoundxpath)
    unknowncompounds = [c for c in thecompounds if
                        getattval(c, 'word') not in compounds.compounds]

    results += unknowncompounds

    # exclude filledpauses, exclude names, misspellings, deviant pronunciations, ......
    allwordnodes = stree.xpath(wordxpath)
    wordnodes = [wn for wn in allwordnodes if
                 len(getattval(wn, 'word')) > 5 and (not isnamenort(wn))]
    unknownwordnodes = [wn for wn in wordnodes if
                        not informlexicon(getattval(wn, 'word').lower())]

    results += unknownwordnodes

    return results


def getunwantedtokens(nortlist: List[Nort]) -> List[Nort]:
    '''
    The function *getunwantedtokens* returns nodes for tokens that are to be discarded
    for sample size as defined in the constant *unwantedtokenlist*

       * .. autodata:: dedup::unwantedtokenlist

    '''
    results = []
    for nort in nortlist:
        nortword = getword(nort)
        if nortword in unwantedtokenlist:
            results.append(nort)
    return results


def samplesize(stree: SynTree) -> List[SynTree]:
    '''
    The function *samplesize* yields the tokens to be excluded from the samplesize. It does so by applying the function
    *samplesize2* and ignoring the DupInfo object that is returned in the tuple.

    .. autofunction:: dedup::samplesize2

    '''
    result, _ = samplesize2(stree)
    return result


def samplesize2(stree: SynTree) -> Tuple[List[SynTree], DupInfo]:
    '''
    The function *samplesize2* yields the tokens to be excluded from the samplesize
    (based on ASTA4 eVersie sec 3, p. 7-8)
    plus a DupInfo object containing a duplicate mapping with items word in pos x is a
    repeat of the     word in position y.

    It maintains two variables for lists of nodes for tokens:

    * **tokennodelist**: the list of nodes for tokens that should be kept, (in their
    original  order), initially all tokens from the yield of *stree*
    * **resultlist**: a list of nodes that should be excluded

      **Remark** there is also a variable *excludednodes*, whose function is not fully
    clear. it might be redundant

    * The function first adds nodes to the *resultlist* that have been found by the *reduce* function in the correction module and that are represented in the metadata.
    * It next adds nodes for symbols that should be discarded. It obtains these nodes via the function *getunwantedtokens*

        * .. autofunction:: dedup::getunwantedtokens


    * It adds nodes for interjections and filledpauses to the resultlist via the function *getfilledpauses*

        * .. autofunction:: dedup::getfilledpauses

    * It adds duplicates of the words *ja*, *nee*, *nou*
    * It adds short repetitions in the tokenlist with *ja*, *nee*, *nou* removed by applying the function *getprefixwords2*

        * .. autofunction:: dedup::getprefixwords2

    '''

    resultlist = []
    alldupinfo = DupInfo()
    # get the token nodes in sequence
    originaltokennodelist = getnodeyield(stree)
    tokennodelist = originaltokennodelist
    excludednodes = []
    # hitprint(tokennodelist)

    # add results that have been found earlier by reduce in correction and that are now in the metadata
    # and exclude these nodes for further processing
    mdnodes = []
    mlumds = stree.xpath(xmetaxpath)
    for mlumd in mlumds:
        if 'value' in mlumd.attrib and mlumd.attrib['value'] in samplesizemdvalues:
            tokenbeginstr = mlumd.attrib['annotatedposlist']
            tokenbegins = string2list(tokenbeginstr)
            for tokenbegin in tokenbegins:
                nodexpath = './/node[@pt and @begin="{}"]'.format(tokenbegin)
                newnode = find1(stree, nodexpath)
                if newnode is not None:
                    mdnodes.append(newnode)
                else:
                    settings.LOGGER.error(
                        'Metadata node not found in tree: md.begin={}'.format(tokenbegin))
                    etree.dump(stree)
    excludednodes += mdnodes
    tokennodelist = [n for n in tokennodelist if n not in excludednodes]
    resultlist += mdnodes

    # throw out unwanted symbols - -- # etc
    unwantedtokens = getunwantedtokens(tokennodelist)
    resultlist += unwantedtokens
    tokennodelist = [n for n in tokennodelist if n not in unwantedtokens]

    # find filledpauses and interjections
    filledpausenodes = getfilledpauses(tokennodelist)
    resultlist += filledpausenodes
    tokennodelist = [n for n in tokennodelist if n not in filledpausenodes]

    # find duplicatenode repetitions of ja, nee, nou
    janeenouduplicatenodes, dupinfo = find_janeenouduplicates2(tokennodelist)
    resultlist += janeenouduplicatenodes
    tokennodelist = [
        n for n in tokennodelist if n not in janeenouduplicatenodes]
    alldupinfo = alldupinfo.merge(dupinfo)

    # temporarily remove ja nee nou to get the right short repetitions
    janeenoutokens = findjaneenou(tokennodelist)
    temptokennodelist = [n for n in tokennodelist if n not in janeenoutokens]

    # find prefix herhalingen < 50%
    def cond(x, y):
        return len(cleanwordofnort(x)) / len(cleanwordofnort(y)) <= 0.5

    prefixnodes, dupinfo = getprefixwords2(temptokennodelist, cond)
    resultlist += prefixnodes
    tokennodelist = [n for n in tokennodelist if n not in prefixnodes]
    alldupinfo = alldupinfo.merge(dupinfo)

    return resultlist, alldupinfo


# initialize filledpauseslexicon
filledpauseslexicon = set()
filledpausesfilename = os.path.join(settings.SD_DIR, 'data', 'filledpauseslexicon',
                                    'filledpauseslexicon.txt')
filledpausesfile = open(filledpausesfilename, 'r', encoding='utf8')
for word in filledpausesfile:
    cleanword = word.strip()
    filledpauseslexicon.add(cleanword)



'''


The namepartlexicon module:

* reads in the data from the file names/nameparts/namepartslexicon.csv
* stores it in the namepartlexicon  dictionary (Dict[str, int]) with its frequency
* provides  functions for checking membership of the namepartlexicon:



.. autofunction:: namepart_isa_namepart
.. autofunction:: namepart_isa_namepart_uc

'''

import csv
import os
from typing import Dict

from sastadev.conf import settings

tab: str = '\t'
namepartlexicon: Dict[str, int] = {}


def namepart_isa_namepart(word: str) -> bool:
    '''
    The function namepart_isa_namepart checks whether the string *word* occurs in the namepartlexicon dictionary.

    '''
    return word in namepartlexicon


def namepart_isa_namepart_uc(word: str) -> bool:
    '''
    The function namepart_isa_namepart_uc checks whether the string *word* with its initial character in upper case
    occurs in the namepartlexicon dictionary.

    '''
    if word is None or word == '':
        result = False
    else:
        uc = word[0].isupper()
        found = word in namepartlexicon
        result = uc and found
    return result


namepartfilename = os.path.join(settings.SD_DIR, 'data', 'names', 'nameparts', 'namepartlexicon.csv')
with open(namepartfilename, 'r', encoding='utf8') as namepartfile:
    csvreader = csv.reader(namepartfile, delimiter=tab)
    for row in csvreader:
        namepart = row[0]
        frq = int(row[1])
        namepartlexicon[namepart] = frq



from collections import Counter
from copy import copy

from lxml import etree

from sastadev.conf import settings
from sastadev.treebankfunctions import getattval, getmarkedyield, getyield

tab = '\t'
space = ' '
eps = ''

usercommentbegin = 0
usercommentuntil = 3
usercommentdefaultvalue = eps


def getmarkedutt(m, syntree):
    thewordlist = getyield(syntree)
    thepositions = getwordpositions(m, syntree)
    themarkedyield = getmarkedyield(thewordlist, thepositions)
    yieldstr = space.join(themarkedyield)
    return yieldstr


def mark(str):
    result = '*' + str + '*'
    return result


def getwordpositionsold(matchtree, syntree):
    positions1 = []
    for node in matchtree.iter():
        if 'pt' in node.attrib:
            if 'end' in node.attrib:
                positions1.append(node.attrib['end'])

    indexednodes = {}
    for node in syntree.iter():
        if 'index' in node.attrib and ('pt' in node.attrib or 'cat' in node.attrib or 'pos' in node.attrib):
            theindex = node.attrib['index']
            indexednodes[theindex] = node

    thequery2 = ".//node[@index and not(@pt) and not(@cat)]"
    try:
        matches2 = matchtree.xpath(thequery2)
    except etree.XPathEvalError as e:
        matches2 = []
    positions2 = []
    for m in matches2:
        positions2 += getwordpositions(m, syntree)
    positions = positions1 + positions2
    result = [int(p) for p in positions]
    return result


def getwordpositions(matchtree, syntree):
    # nothing special needs to be done for index nodes since they also have begin and end
    positions = []
    for node in matchtree.iter():
        if 'end' in node.attrib:
            positions.append(node.attrib['end'])
    result = [int(p) for p in positions]
    return result


def getfirstwordposition(matchtree):
    if 'begin' in matchtree.attrib:
        positionstr = getattval(matchtree, 'begin')
        position = int(positionstr) + 1
    else:
        position = 0
    return position

#moved to treebannkfunctions
# def getmarkedyield(wordlist, positions):
#     pos = 1
#     resultlist = []
#     for w in wordlist:
#         if pos in positions:
#             resultlist.append(mark(w))
#         else:
#             resultlist.append(w)
#         pos += 1
#     return resultlist


def mismatches(queryid, queries, theresultsminusgold, goldminustheresults, allmatches, allutts, platinumcheckfile):
    if theresultsminusgold != {}:
        print('More examples', file=platinumcheckfile)
    for uttid in theresultsminusgold:
        if (queryid, uttid) in allmatches:
            for (m, syntree) in allmatches[(queryid, uttid)]:
                markedutt = getmarkedutt(m, syntree)
                platinumcheckrow1 = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item,
                                     uttid, markedutt]
                print(tab.join(platinumcheckrow1), file=platinumcheckfile)

    if goldminustheresults != {}:
        print('Missed examples', file=platinumcheckfile)
    for uttid in goldminustheresults:
        if uttid in allutts:
            uttstr = space.join(allutts[uttid])
        else:
            settings.LOGGER.warning('uttid {} not in alluts'.format(uttid))
        platinumcheckrow2 = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item, uttid,
                             uttstr]
        print(tab.join(platinumcheckrow2), file=platinumcheckfile)


def getmarkposition(position, nodeendmap, uttid):
    if position == 0:
        result = 1
    elif uttid in nodeendmap:
        if str(position) in nodeendmap[uttid]:
            result = nodeendmap[uttid][str(position)]
        else:
            settings.LOGGER.error('getmarkposition: No mapping found for position {} in utterance {}'.format(position, uttid))
            result = 1
    else:
        settings.LOGGER.error('getmarkposition: No mappings found for uttid {}'.format(uttid))
        result = 1
    return result


def exactmismatches(queryid, queries, exactresults, exactgoldscores, allmatches, allutts, platinumcheckfile,
                    permsilverdatadict={}, annotationinput=False):
    theexactresults = exactresults[queryid] if queryid in exactresults else Counter()
    theexactgoldscores = exactgoldscores[queryid] if queryid in exactgoldscores else Counter()
    (theresultsminusgold, goldminustheresults, intersection) = exactcompare(theexactresults, theexactgoldscores)
    newrows = []
    if theresultsminusgold != []:
        print('More examples', file=platinumcheckfile)
    for hit in theresultsminusgold:
        uttid, position = hit
        if (queryid, uttid) in allmatches or annotationinput:
            #markposition = 1 if position == 0 else position
            markposition = position
            markedwordlist = getmarkedyield(allutts[uttid], [markposition])
            uttstr = space.join(markedwordlist)
            platinumcheckrow1 = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item,
                                 str(uttid), str(markposition), uttstr]
            print(tab.join(platinumcheckrow1), file=platinumcheckfile)
            key = (queryid, uttid, position)
            usercomments = getusercomments(permsilverdatadict, key, report=True)
            xlplatinumcheckrow1 = usercomments + ['More examples'] + platinumcheckrow1
            newrows.append(xlplatinumcheckrow1)
            # for (m, syntree) in allmatches[(queryid, uttid)]:
            #    if getfirstwordposition(m) == position:
            #        markedutt = getmarkedutt(m, syntree)
            #       platinumcheckrow1 = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item,
            #                            uttid), markedutt]
            #        print(tab.join(platinumcheckrow1), file=platinumcheckfile)

    if goldminustheresults != []:
        print('Missed examples', file=platinumcheckfile)
    for hit in goldminustheresults:
        (uttid, position) = hit
        if uttid in allutts:
            #markposition = 1 if position == 0 else position
            markposition = position
            markedwordlist = getmarkedyield(allutts[uttid], [markposition])
            uttstr = space.join(markedwordlist)
        else:
            settings.LOGGER.warning('uttid {} not in allutts'.format(uttid))
            uttstr = ""
            markposition = 0
        platinumcheckrow2 = [queryid, queries[queryid].cat, queries[queryid].subcat, queries[queryid].item, str(uttid),
                             str(markposition),
                             uttstr]
        print(tab.join(platinumcheckrow2), file=platinumcheckfile)
        key = (queryid, uttid, position)
        usercomments = getusercomments(permsilverdatadict, key, report=False)
        xlplatinumcheckrow2 = usercomments + ['Missed examples'] + platinumcheckrow2
        newrows.append(xlplatinumcheckrow2)
    return newrows


def compareunaligned(resultctr, goldctr):
    '''

    :param resultlist:
    :param goldlist:
    :return:
    '''
    resultlist = counter2list(resultctr)
    goldlist = counter2list(goldctr)
    curgoldlist = copy(goldlist)
    newintersection = []
    takefromresultlist = []
    takefromgoldlist = []
    for (utt1, pos1) in resultlist:
        if (utt1, 0) in curgoldlist:
            takefromresultlist.append((utt1, pos1))
            takefromgoldlist.append((utt1, 0))
            newintersection.append((utt1, pos1))
            curgoldlist.remove((utt1, 0))
        elif pos1 == 0:
            for (utt2, pos2) in curgoldlist:
                if utt1 == utt2:
                    takefromresultlist.append((utt1, pos1))
                    takefromgoldlist.append((utt1, pos2))
                    newintersection.append((utt1, pos2))
                    curgoldlist.remove((utt2, pos2))
                    break
    takefromresultctr = Counter(takefromresultlist)
    takefromgoldctr = Counter(takefromgoldlist)
    newintersectionctr = Counter(newintersection)
    return (takefromresultctr, takefromgoldctr, newintersectionctr)


def exactcompare(exactresults, exactgoldscores):
    '''
    compares two lists of exact results, i.e. dlists of pairs (uttid, position)
    :param exactresults:
    :param exactgoldscores:
    :return: triple (resultsminusgold, goldminusresults, intereection)
    '''
    resultscounter = Counter(exactresults)
    goldcounter = Counter(exactgoldscores)
    intersection1 = resultscounter & goldcounter
    resultsminusgold1 = resultscounter - goldcounter
    goldminusresults1 = goldcounter - resultscounter

    (resultsminusgold2, goldminusresults2, intersection2) = compareunaligned(resultsminusgold1, goldminusresults1)

    intersectionctr = intersection1 + intersection2
    resultsminusgoldctr = resultsminusgold1 - resultsminusgold2
    goldminusresultsctr = goldminusresults1 - goldminusresults2

    intersection = counter2list(intersectionctr)
    resultsminusgold = counter2list(resultsminusgoldctr)
    goldminusresults = counter2list(goldminusresultsctr)

    return (resultsminusgold, goldminusresults, intersection)


def counter2list(ctr):
    result1 = [el for el in ctr for k in range(ctr[el])]
    result = sorted(result1)
    return result


def getusercomments(permsilverdict, key, report=False):
    if key in permsilverdict:
        therow = permsilverdict[key]
        usercomments = therow[usercommentbegin:usercommentuntil]
        result = usercomments
    else:
        count = usercommentuntil - usercommentbegin
        resultlist = [usercommentdefaultvalue for _ in range(count)]
        result = resultlist
        if report:
            settings.LOGGER.warning('No silver remark for key: {}'.format(key))
    return result


def testcompare():
    testresults = [(1, 2), (1, 2), (1, 2), (1, 5), (1, 6), (2, 0), (2, 4)]
    goldresults = [(1, 2), (2, 4), (2, 6), (1, 0), (3, 5)]
    reftestminusgold = [(1, 2), (1, 5), (1, 6)]
    refgoldminustest = [(3, 5)]
    refintersection = [(1, 2), (1, 2), (2, 4), (2, 6)]
    (testminusgold, goldminustest, intersection) = exactcompare(testresults, goldresults)
    for (l, r, g) in zip(['R-G', 'G-R', 'R*G'], [testminusgold, goldminustest, intersection],
                         [reftestminusgold, refgoldminustest, refintersection]):
        if r == g:
            print('{}: OK {} == {}'.format(l, r, g))
        else:
            print('{}: NO: {} != {}'.format(l, r, g))


if __name__ == '__main__':
    testcompare()



from typing import List

from sastadev.sastatypes import SynTree
from sastadev.treebankfunctions import clausecats, getattval

conjxpath = './/node[@cat="conj" and (count(node)=3 or count(node)=5)]'

#//node[@cat="conj" and (count(node)=3 or count(node)=5) and
#node[@rel="crd" and (@lemma="en" or @lemma="of")] and
#node[@rel="cnj" and @cat!="smain" and @cat!="sv1" and @cat!="whq"]
#]


def getptcat(node):
    pt = getattval(node, 'pt')
    cat = getattval(node, 'cat')
    if pt != '':
        result = pt
    elif cat != '':
        result = cat
    else:
        result = None
    return result


def xenx(tree: SynTree) -> List[SynTree]:
    '''
    The function *xenx* selects conjuncts consisting of 3 or 5 children: either two
    conjuncts and one coordinator, or three conjuncts and 2 coordinators. The conjuncts
    must have the same values for *pt*  or *cat* attributes.
    '''
    results = []
    conjs = tree.xpath(conjxpath)
    #print(conjs)
    for conj in conjs:
        include = True
        firstptcat = None
        crdresult = None
        for child in conj:
            #print('include=', include)
            #print('child', getyield(child))
            childrel = getattval(child, 'rel')
            childptcat = getptcat(child)
            childlemma = getattval(child, 'lemma')
            #print(childrel, childptcat, firstptcat, '<'+childlemma+'>')
            if include:
                if childrel == 'crd':
                    include = include and (childlemma in ['en', 'of'])
                    crdresult = child
                elif childrel == 'cnj':
                    if childptcat is None:
                        include = False
                    if firstptcat is None:
                        firstptcat = childptcat
                    include = include and (childptcat == firstptcat) and childptcat not in clausecats
            #print('include=', include)
        #print('final: include=', include)
        if include:
            results.append(crdresult)
    return results



from sastadev.macros import expandmacros, macrodict

resultlist = []
for macroname, macrovalue in macrodict.items():
    expandedmacrovalue = expandmacros(macrovalue)
    resultlist.append((macroname, len(expandedmacrovalue)))

querylist = ['//node[( %declarative% )]', '//node[( %Ond%  )',
             '//node[(   %Tarsp_B_X_count% = 3)]', '//node[%declarative% and %Ond%]',
             '//node[%declarative% and %Tarsp_B_X_count% = 3]']
#Invalid query: unknown error

queryresultlist = []
for query in querylist:
    expandedquery = expandmacros(query)
    queryresultlist.append((query, len(expandedquery)))

print('\n\nQueries:\n')
sortedqueryresultlist = sorted(queryresultlist, key=lambda x: x[1], reverse=True)
for (query, lvalue) in sortedqueryresultlist:
    print(f'{query}\t{lvalue}')

print('\n\nMacros:\n')
sortedresults = sorted(resultlist, key=lambda x: x[1], reverse=True)
for (name, lvalue) in sortedresults:
    print(f'{name}\t{lvalue}')



"""
The module *tblex* contains functiond that require function
from the lexicon module and from the treebankfunctions module
"""

import sastadev.lexicon as lex
from sastadev.sastatypes import SynTree
from sastadev.treebankfunctions import (all_lower_consonantsnode, getattval,
                                        is_duplicate_spec_noun, iscompound,
                                        isdiminutive, isnumber,
                                        issubstantivised_verb, sasta_long,
                                        sasta_pseudonym, short_nucl_n,
                                        spec_noun)


def recognised_wordnodepos(node: SynTree, pos: str) -> bool:
    '''
    The function *recognised_wordnodepos* determines for *node* whether it is a known
    word of part of speech code *pos*.

    It distinguishes several subcases that yield the result True:

    * the value of the *word* attribute of *node* is a known word form (as determined by the function *lex.informlexiconpos*

    * the lower-cased value of the *word* attribute of *node* is a known word form (as determined by the function *lex.informlexiconpos*

    * the node is a node for a compound, as determined by the function *iscompound*:

        .. autofunction:: treebankfunctions::iscompound
           :noindex:

    * the node is a node for a diminutive, as determined by the function *isdiminutive*:

        .. autofunction:: treebankfunctions::isdiminutive
           :noindex:

    * the node is a node for a name part, as determined by the function *lex.isa_namepart*


    '''
    word = getattval(node, 'word')
    lcword = word.lower()
    result = lex.informlexiconpos(word, pos) or lex.informlexiconpos(lcword, pos) or \
        iscompound(node) or isdiminutive(node) or lex.isa_namepart_uc(word)
    return result


def recognised_wordnode(node: SynTree) -> bool:
    '''
    The function *recognised_wordnode* determines for *node* whether it is a known word.

    It distinguishes several subcases that yield the result True:

    * the value of the *word* attribute of *node* is a known word form (as determined
    by the function *lex.informlexicon*

    * the lower-cased value of the *word* attribute of *node* is a known word form (as
    determined by the function *lex.informlexicon

    * the node is a node for a compound, as determined by the function *iscompound*:

        .. autofunction:: treebankfunctions::iscompound

    * the node is node for a diminutive, as determined by the function *isdiminutive*:

        .. autofunction:: treebankfunctions::isdiminutive

    * the node is a node for a name part, as determined by the function *lex.isa_namepart*


    '''

    word = getattval(node, 'word')
    lcword = word.lower()
    result = lex.informlexicon(word) \
        or lex.informlexicon(lcword) \
        or iscompound(node) \
        or isdiminutive(node) \
        or lex.isa_namepart(word)
    return result


def recognised_lemmanode(node: SynTree) -> bool:
    '''
    The function *recognised_lemmanode* checks whether the *lemma* of *node* is in
    the lexicon  (as determined by the function *lex.informlexicon*).

    '''
    lemma = getattval(node, 'lemma')
    result = lex.informlexicon(lemma)
    return result


def recognised_lemmanodepos(node: SynTree, pos: str) -> bool:
    '''
    The function *recognised_lemmanodepos* checks whether the *lemma* of *node* is in
    the lexicon with part of speech *pos* (as determined by * lex.informlexiconpos*).

    '''
    lemma = getattval(node, 'lemma')
    result = lex.informlexiconpos(lemma, pos)
    return result


def asta_recognised_lexnode(node: SynTree) -> bool:
    '''
    The function *asta_recognised_lexnode* determines whether *node* should count as a
    lexical verb in the ASTA method.

    This is the case if *pt* equals *ww* and the node is not a substantivised verb as
    determined by the function *issubstantivised_verb*:

    .. autofunction:: treebankfunctions::issubstantivised_verb

    '''
    if issubstantivised_verb(node):
        result = False
    else:
        result = getattval(node, 'pt') == 'ww'
    return result


def asta_recognised_nounnode(node: SynTree) -> bool:
    '''
    The function *asta_recognised_nounnode* determines whether *node* should count as a
    noun in the ASTA method.

    This is the case if

    * either the node meets the conditions of *sasta_pseudonym*

       .. autofunction:: treebankfunctions::sasta_pseudonym

    * or the node meets the conditions of *spec_noun*

       .. autofunction:: treebankfunctions::spec_noun

    * or the node meets the conditions of *is_duplicate_spec_noun*

       .. autofunction:: treebankfunctions::is_duplicate_spec_noun

    * or the node meets the conditions of *sasta_long*

       .. autofunction:: treebankfunctions::sasta_long

    * or the node meets the conditions of *recognised_wordnodepos*

       .. autofunction:: treebankfunctions::recognised_wordnodepos

    * or the node meets the conditions of *recognised_lemmanodepos(node, pos)*

       .. autofunction:: treebankfunctions::recognised_lemmanodepos(node, pos)

    However, the node should:

    * neither consist of lower case consonants only, as determined by *all_lower_consonantsnode*:

       .. autofunction:: treebankfunctions::all_lower_consonantsnode

    * nor satisfy the conditions of *short_nucl_n*:

       .. autofunction:: treebankfunctions::short_nucl_n

    '''

    if issubstantivised_verb(node):
        pos = 'ww'
    else:
        pos = 'n'
    result = sasta_pseudonym(node)
    result = result or spec_noun(node)
    result = result or is_duplicate_spec_noun(node)
    result = result or sasta_long(node)
    result = result or recognised_wordnodepos(node, pos)
    result = result or recognised_lemmanodepos(node, pos)
    result = result and not (all_lower_consonantsnode(node))
    result = result and not (short_nucl_n(node))
    return result


def asta_recognised_wordnode(node: SynTree) -> bool:
    result = sasta_pseudonym(node)
    result = result or spec_noun(node)
    result = result or is_duplicate_spec_noun(node)
    result = result or sasta_long(node)
    result = result or recognised_wordnode(node)
    result = result or recognised_lemmanode(node)
    result = result or isnumber(node)
    result = result and not (all_lower_consonantsnode(node))
    result = result and not (short_nucl_n(node))
    return result



from lxml import etree

from sastadev.ASTApostfunctions import neologisme, phonpar, sempar


def test(stree):
    neoresults = neologisme(stree)
    semparresults = sempar(stree)
    phonparresults = phonpar(stree)
    results = [('neo', neoresult) for neoresult in neoresults] +\
              [('sempar', semparresult) for semparresult in semparresults] +\
              [('phonpar', phonparresult) for phonparresult in phonparresults]
    return results


def main():
    for i in strees:
        results = test(strees[i])
        for result in results:
            print('{}: {}:{}'.format(result[0], result[1].attrib['word'], result[1].attrib['begin']))


streestrings = {}

streestrings[1] = """
<alpino_ds version="1.3">
  <metadata><meta name="uttid" value="1" type="text"/><meta name="xsid" value="1" type="text"/><meta name="origutt" value="ik heb geduusterd [*n]" type="text"/><xmeta annotatedposlist="[2]" annotatedwordlist="['geduusterd']" annotationposlist="[4]" annotationwordlist="['n']" atype="text" backplacement="0" cat="None" name="Error Marking" penalty="10" source="CHAT" subcat="None" value="['n']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'geduusterd', '[*', 'n', ']']" atype="list" backplacement="0" cat="None" name="tokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'geduusterd', '[*', 'n', ']']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'geduusterd']" atype="list" backplacement="0" cat="None" name="cleanedtokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'geduusterd']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="[0, 1, 2]" atype="list" backplacement="0" cat="None" name="cleanedtokenpositions" penalty="10" source="CHAT/Tokenisation" subcat="None" value="[0, 1, 2]"/></metadata><node begin="0" cat="top" end="3" id="0" rel="top">
    <node begin="0" cat="smain" end="3" id="1" rel="--">
      <node begin="0" case="nom" def="def" end="1" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" id="2" index="1" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
      <node begin="1" end="2" frame="verb(hebben,sg1,aux_psp_hebben)" id="3" infl="sg1" lcat="smain" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="aux_psp_hebben" sense="heb" stype="declarative" tense="present" word="heb" wvorm="pv"/>
      <node begin="0" cat="ppart" end="3" id="4" rel="vc">
        <node begin="0" end="1" id="5" index="1" rel="su"/>
        <node begin="2" buiging="zonder" end="3" frame="verb('hebben/zijn',psp,intransitive)" id="6" infl="psp" lcat="ppart" lemma="geduusterd" pos="verb" positie="vrij" postag="WW(vd,vrij,zonder)" pt="ww" rel="hd" root="geduusterd" sc="intransitive" sense="geduusterd" word="geduusterd" wvorm="vd"/>
      </node>
    </node>
  </node>
  <sentence sentid="1">ik heb geduusterd</sentence>
  <comments>
    <comment>Q#ng1646152422|ik heb geduusterd|1|1|-5.158487943820001</comment>
  </comments>
</alpino_ds>
"""


streestrings[2] = """
  <alpino_ds version="1.3">
  <metadata><meta name="uttid" value="1" type="text"/><meta name="xsid" value="1" type="text"/><meta name="origutt" value="ik heb ngeduusterd [*n]" type="text"/><xmeta annotatedposlist="[2]" annotatedwordlist="['ngeduusterd']" annotationposlist="[4]" annotationwordlist="['n']" atype="text" backplacement="0" cat="None" name="Error Marking" penalty="10" source="CHAT" subcat="None" value="['n']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'ngeduusterd', '[*', 'n', ']']" atype="list" backplacement="0" cat="None" name="tokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'ngeduusterd', '[*', 'n', ']']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'ngeduusterd']" atype="list" backplacement="0" cat="None" name="cleanedtokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'ngeduusterd']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="[0, 1, 2]" atype="list" backplacement="0" cat="None" name="cleanedtokenpositions" penalty="10" source="CHAT/Tokenisation" subcat="None" value="[0, 1, 2]"/></metadata><node begin="0" cat="top" end="3" id="0" rel="top">
    <node begin="0" cat="smain" end="3" id="1" rel="--">
      <node begin="0" case="nom" def="def" end="1" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" id="2" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
      <node begin="1" end="2" frame="verb(hebben,sg1,transitive_ndev)" id="3" infl="sg1" lcat="smain" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="transitive_ndev" sense="heb" stype="declarative" tense="present" word="heb" wvorm="pv"/>
      <node begin="2" end="3" frame="noun(both,both,both)" gen="both" genus="zijd" getal="ev" graad="basis" id="4" lcat="np" lemma="ngeduusterd" naamval="stan" ntype="soort" num="both" pos="noun" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="obj1" rnum="sg" root="ngeduusterd" sense="ngeduusterd" word="ngeduusterd"/>
    </node>
  </node>
  <sentence sentid="1">ik heb ngeduusterd</sentence>
  <comments>
    <comment>Q#ng1646219407|ik heb ngeduusterd|1|1|-1.6311900273499995</comment>
  </comments>
</alpino_ds>
"""

streestrings[3] = """
  <alpino_ds version="1.3">
  <metadata><meta name="uttid" value="2" type="text"/><meta name="xsid" value="2" type="text"/><meta name="origutt" value="ik heb nngeduusterd@n" type="text"/><xmeta annotatedposlist="[2]" annotatedwordlist="['nngeduusterd@n']" annotationposlist="[]" annotationwordlist="['@n']" atype="text" backplacement="0" cat="None" name="Special Form" penalty="10" source="CHAT" subcat="None" value="['@n']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'nngeduusterd@n']" atype="list" backplacement="0" cat="None" name="tokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'nngeduusterd@n']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'nngeduusterd']" atype="list" backplacement="0" cat="None" name="cleanedtokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'nngeduusterd']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="[0, 1, 2]" atype="list" backplacement="0" cat="None" name="cleanedtokenpositions" penalty="10" source="CHAT/Tokenisation" subcat="None" value="[0, 1, 2]"/></metadata><node begin="0" cat="top" end="3" id="0" rel="top">
    <node begin="0" cat="smain" end="3" id="1" rel="--">
      <node begin="0" case="nom" def="def" end="1" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" id="2" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
      <node begin="1" end="2" frame="verb(hebben,sg1,transitive_ndev)" id="3" infl="sg1" lcat="smain" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="transitive_ndev" sense="heb" stype="declarative" tense="present" word="heb" wvorm="pv"/>
      <node begin="2" end="3" frame="noun(both,both,both)" gen="both" genus="zijd" getal="ev" graad="basis" id="4" lcat="np" lemma="nngeduusterd" naamval="stan" ntype="soort" num="both" pos="noun" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="obj1" rnum="sg" root="nngeduusterd" sense="nngeduusterd" word="nngeduusterd"/>
    </node>
  </node>
  <sentence sentid="2">ik heb nngeduusterd</sentence>
  <comments>
    <comment>Q#ng1646219408|ik heb nngeduusterd|1|1|-1.6311900273499995</comment>
  </comments>
</alpino_ds>
"""
streestrings[4] = """
  <alpino_ds version="1.3">
  <metadata><meta name="uttid" value="3" type="text"/><meta name="xsid" value="3" type="text"/><meta name="origutt" value="ik heb pgeduusterd [*p]" type="text"/><xmeta annotatedposlist="[2]" annotatedwordlist="['pgeduusterd']" annotationposlist="[4]" annotationwordlist="['p']" atype="text" backplacement="0" cat="None" name="Error Marking" penalty="10" source="CHAT" subcat="None" value="['p']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'pgeduusterd', '[*', 'p', ']']" atype="list" backplacement="0" cat="None" name="tokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'pgeduusterd', '[*', 'p', ']']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'pgeduusterd']" atype="list" backplacement="0" cat="None" name="cleanedtokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'pgeduusterd']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="[0, 1, 2]" atype="list" backplacement="0" cat="None" name="cleanedtokenpositions" penalty="10" source="CHAT/Tokenisation" subcat="None" value="[0, 1, 2]"/></metadata><node begin="0" cat="top" end="3" id="0" rel="top">
    <node begin="0" cat="smain" end="3" id="1" rel="--">
      <node begin="0" case="nom" def="def" end="1" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" id="2" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
      <node begin="1" end="2" frame="verb(hebben,sg1,transitive_ndev)" id="3" infl="sg1" lcat="smain" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="transitive_ndev" sense="heb" stype="declarative" tense="present" word="heb" wvorm="pv"/>
      <node begin="2" end="3" frame="noun(both,both,both)" gen="both" genus="zijd" getal="ev" graad="basis" id="4" lcat="np" lemma="pgeduusterd" naamval="stan" ntype="soort" num="both" pos="noun" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="obj1" rnum="sg" root="pgeduusterd" sense="pgeduusterd" word="pgeduusterd"/>
    </node>
  </node>
  <sentence sentid="3">ik heb pgeduusterd</sentence>
  <comments>
    <comment>Q#ng1646219409|ik heb pgeduusterd|1|1|-1.6311900273499995</comment>
  </comments>
</alpino_ds>
"""

streestrings[5] = """
  <alpino_ds version="1.3">
  <metadata><meta name="uttid" value="4" type="text"/><meta name="xsid" value="4" type="text"/><meta name="origutt" value="ik heb sgeduusterd [*s]" type="text"/><xmeta annotatedposlist="[2]" annotatedwordlist="['sgeduusterd']" annotationposlist="[4]" annotationwordlist="['s']" atype="text" backplacement="0" cat="None" name="Error Marking" penalty="10" source="CHAT" subcat="None" value="['s']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'sgeduusterd', '[*', 's', ']']" atype="list" backplacement="0" cat="None" name="tokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'sgeduusterd', '[*', 's', ']']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="['ik', 'heb', 'sgeduusterd']" atype="list" backplacement="0" cat="None" name="cleanedtokenisation" penalty="10" source="CHAT/Tokenisation" subcat="None" value="['ik', 'heb', 'sgeduusterd']"/><xmeta annotatedposlist="[]" annotatedwordlist="[]" annotationposlist="[]" annotationwordlist="[0, 1, 2]" atype="list" backplacement="0" cat="None" name="cleanedtokenpositions" penalty="10" source="CHAT/Tokenisation" subcat="None" value="[0, 1, 2]"/></metadata><node begin="0" cat="top" end="3" id="0" rel="top">
    <node begin="0" cat="smain" end="3" id="1" rel="--">
      <node begin="0" case="nom" def="def" end="1" frame="pronoun(nwh,fir,sg,de,nom,def)" gen="de" getal="ev" id="2" lcat="np" lemma="ik" naamval="nomin" num="sg" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,ev)" pt="vnw" rel="su" rnum="sg" root="ik" sense="ik" status="vol" vwtype="pers" wh="nwh" word="ik"/>
      <node begin="1" end="2" frame="verb(hebben,sg1,transitive_ndev)" id="3" infl="sg1" lcat="smain" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="transitive_ndev" sense="heb" stype="declarative" tense="present" word="heb" wvorm="pv"/>
      <node begin="2" end="3" frame="noun(both,both,both)" gen="both" genus="zijd" getal="ev" graad="basis" id="4" lcat="np" lemma="sgeduusterd" naamval="stan" ntype="soort" num="both" pos="noun" postag="N(soort,ev,basis,zijd,stan)" pt="n" rel="obj1" rnum="sg" root="sgeduusterd" sense="sgeduusterd" word="sgeduusterd"/>
    </node>
  </node>
  <sentence sentid="4">ik heb sgeduusterd</sentence>
  <comments>
    <comment>Q#ng1646219410|ik heb sgeduusterd|1|1|-1.6311900273499995</comment>
  </comments>
</alpino_ds>

"""

strees = {}
for i in streestrings:
    strees[i] = etree.fromstring(streestrings[i])

if __name__ == '__main__':
    main()



import os
import sys

reportwidth = 0

# the following 2 are useful if you just have to read the files,
# they are less useful if you have to create new copies with the same folder structure


def getallfilenames(inpath, allowedexts):
    filenames = []
    for root, dirs, thefiles in os.walk(inpath):
        for filename in thefiles:
            fullname = os.path.join(root, filename)
            (base, ext) = os.path.splitext(filename)
            if ext in allowedexts:
                filenames.append(fullname)
    return filenames


def iterallfilenames(inpath, allowedexts):
    filenames = []
    for root, dirs, thefiles in os.walk(inpath):
        for filename in thefiles:
            fullname = os.path.join(root, filename)
            (base, ext) = os.path.splitext(filename)
            if ext in allowedexts:
                yield fullname


# @@add functions that arev useful if you have make new versions with the same folder structure

def reportevery(ctr, repevery, sep=' ', maxwidth=100, file=sys.stderr):
    global reportwidth
    # curreportwidth = reportwidth
    if ctr % repevery == 0:
        ctrstr = str(ctr)
        lcurreport = len(ctrstr) + len(sep)
        if reportwidth + lcurreport > maxwidth:
            print(file=sys.stderr)
            reportwidth = lcurreport
        else:
            print(ctr, end=sep, file=file)
            sys.stderr.flush()
            reportwidth += len(str(ctr)) + len(sep)


def getbasename(fullname):
    path, filename = os.path.split(fullname)
    base, ext = os.path.splitext(filename)
    return base



'''
The CELEX lexicon consists of 3 backslash-separated value files:

* DMW (dmwcdok.txt): word forms and their properties
* DML (dmlcd.txt): lemmas and their morphological properties
* DSL (dslcd.txt): lemmas and their syntactic properties

These files can be found in the folder celexlexicondata/dutch in the code folder.

We store these in the celexlexicon module in python dictionaries:

* dmwkeydict: key = word identifier, value = whole record for this identifier
* dmldict: key = lemma identifier, value = whole record for this identifier
* dsldict: key = lemma identifier, value = whole dsl record for this identifier

We define the following indexes (as python dictionaries) for quick lookup in this module:

* dmwlemmakeyinflindex: key = (lemma identifier, infl), value = List of word form identifiers
* dmwdict: key = word form, value = list of word identifiers

This module also contains functions to map CELEX inflection codes to DCOI inflection codes, and the other way around,
as well as functions to generate inflected forms for a given lemma, pos-code and inflectional properties.

'''

import csv
import os
import re
import sys
from collections import defaultdict
from typing import Dict, List, Optional, Tuple

from sastadev import treebankfunctions
from sastadev.conf import settings
from sastadev.sastatypes import CELEX_INFL, SynTree, WordInfo

backslash = '\\'
celexsep = backslash

pospattern = r'^.*\[(?P<pos>.)\]$'
posre = re.compile(pospattern)


# dml columns
IdNum, Head, Inl, MorphStatus, MorphCnt, DerComp, Comp, Def, Imm, \
    ImmSubCat, ImmAllo, ImmSubst, StrucLab, StrucAllo, StrucSubst, Sepa = 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15

# verbs that do not necessarily have a t suffix  in present tense singular
no_t_verbs = {'mogen', 'kunnen', 'zullen', 'willen'}

logfile = sys.stderr

# initialisation
# read the celex lexicon
inputfolder = os.path.join(settings.SD_DIR, 'data', 'celexlexicon', 'dutch')

dmwfilename = 'DMWCDOK.txt'
dmwfullname = os.path.join(inputfolder, dmwfilename)
dmwdict: Dict[str, List[str]] = {}
dmwlemmakeyinflindex: Dict[Tuple[str, str], List[str]] = defaultdict(list)
dmwkeydict = {}

# dmw columns: recid, form, lemmakey, pos, infl


with open(dmwfullname, mode='r') as infile:
    myreader = csv.reader(infile, delimiter=celexsep)
    for row in myreader:
        thekey = row[0]
        dmwkeydict[thekey] = row
        theform = row[1]
        if theform in dmwdict:
            dmwdict[theform].append(thekey)
        else:
            dmwdict[theform] = [thekey]
        lemmainflkey = (row[3], row[4])
        dmwlemmakeyinflindex[lemmainflkey].append(row[0])

dmlfilename = 'DMLCD.txt'
dmlfullname = os.path.join(inputfolder, dmlfilename)
dmldict: Dict[str, List[str]] = {}

with open(dmlfullname, mode='r') as infile:
    myreader = csv.reader(infile, delimiter=celexsep)
    for row in myreader:
        thekey = row[IdNum]
        if thekey in dmldict:
            print('Warning: Duplicate key in dmlcd: {}'.format(thekey), file=logfile)
        dmldict[thekey] = row


# The dsl.cd file contains the following fields:
dslIdNum, dslHead, dslInl, dslClassNum, dslGendNum, dslDeHetNum, dslPropNum, \
    dslAuxNum, dslSubClassVNum, dslSubCatNum, dslAdvNum, dslCardOrdNum, dslSubClassPNum = 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12

posnum2pos = {'0': 'None', '1': 'n', '2': 'adj', '3': 'tw', '4': 'ww', '5': 'lid', '6': 'vnw', '7': 'bw', '8': 'vz', '9': 'vw', '10': 'tsw'}
pos2posnum = {posnum2pos[key]: key for key in posnum2pos}


dslfilename = 'DSLCD.txt'
dslfullname = os.path.join(inputfolder, dslfilename)
dsldict: Dict[str, List[str]] = {}
dsllemmaposindex: Dict[Tuple[str, str], List[str]] = defaultdict(list)


with open(dslfullname, mode='r') as infile:
    myreader = csv.reader(infile, delimiter=celexsep)
    for row in myreader:
        thekey = row[dslIdNum]
        if thekey in dsldict:
            print('Warning: Duplicate key in dslcd: {}'.format(thekey), file=logfile)
        dsldict[thekey] = row
        lemmaposkey = (row[dslHead], row[dslClassNum])
        dsllemmaposindex[lemmaposkey].append(thekey)


def dcoiphi2celexpv(thesubj: SynTree, thepv: SynTree, inversion: bool) -> str:
    '''
    :param thesubj: subject node
    :param thepv: finite verb node
    :param inversion: True or False
    :return:
    '''
    (rawperson, number, _) = treebankfunctions.getphi(thesubj)
    lemma = treebankfunctions.getlemma(thesubj)
    if lemma in ['het']:
        person = '3'
        number = 'ev'
    else:
        person = '3' if rawperson == '' else rawperson
    tense = treebankfunctions.getattval(thepv, 'pvtijd')
    pvnumber = treebankfunctions.getattval(thepv, 'pvagr')
    celexperson = person[0] if person != '' and ((number in ['ev', 'getal'] and tense == 'tgw') or (tense == '')) else ''
    if number != 'getal' and number != '':
        celexnumber = number[0]
    elif lemma in ['je']:
        celexnumber = 'e'
    elif celexperson in ['1', '2', '3']:
        celexnumber = 'e'
    elif pvnumber == 'met-t':
        celexnumber = 'e'
    elif pvnumber == '':
        celexnumber = 'm'  # assuming that we are dealing with an infinitive
    else:
        celexnumber = pvnumber[0]
    celextense = tense[0] if tense != '' else 't'  # in case a verb is analysed as inf instead of as  pv
    celexinversion = 'I' if inversion else ''
    result = celextense + celexnumber + celexperson + celexinversion
    return result


celex2dcoimap: Dict[str, Dict[str, str]] =\
    {'te1': {'pvtijd': 'tgw', 'pvagr': 'ev', 'wvorm': 'pv'},
     'te2': {'pvtijd': 'tgw', 'pvagr': 'ev', 'wvorm': 'pv'},
     'te2t': {'pvtijd': 'tgw', 'pvagr': 'met-t', 'wvorm': 'pv'},
     'te3': {'pvtijd': 'tgw', 'pvagr': 'ev', 'wvorm': 'pv'},
     'te3t': {'pvtijd': 'tgw', 'pvagr': 'met-t', 'wvorm': 'pv'},
     'te2I': {'pvtijd': 'tgw', 'pvagr': 'ev', 'wvorm': 'pv'},
     'tm': {'pvtijd': 'tgw', 'pvagr': 'mv', 'wvorm': 'pv'},
     've': {'pvtijd': 'verl', 'pvagr': 'ev', 'wvorm': 'pv'},
     'vm': {'pvtijd': 'verl', 'pvagr': 'mv', 'wvorm': 'pv'},
     'i': {'wvorm': 'inf', 'positie': 'vrij', 'buiging': 'zonder'},
     'pv': {'wvorm': 'vd', 'positie': 'vrij', 'buiging': 'zonder'},
     'pt': {'wvorm': 'td', 'positie': 'vrij', 'buiging': 'zonder'},
     'pvE': {'wvorm': 'vd', 'buiging': 'met-e', 'positie': 'prenom'},
     'ptE': {'wvorm': 'td', 'buiging': 'met-e', 'positie': 'prenom'}
     }


def celexpv2dcoi(word: str, infl: str, lemma: str) -> Dict[str, str]:
    results: Dict[str, str] = {}
    if infl not in celex2dcoimap:
        results = {}
    elif infl in {'te2', 'te3'}:
        if lemma[-3] == 't':
            results = celex2dcoimap[infl]
        elif lemma in no_t_verbs:
            results = celex2dcoimap[infl]
        else:
            results = celex2dcoimap[infl + 't']
    else:
        results = celex2dcoimap[infl]
    return results


def incelexdmw(str: str) -> bool:
    result = str in dmwdict
    return result


def incelexdmwpos(word: str, pos: str) -> bool:
    result = incelexdmw(word)
    poslist = getposlist(word)
    result = result and pos in poslist
    return result


def getlemmas(word: str) -> List[str]:
    '''
    yields a list of lemmas for the input string word
    :param word:
    :return:
    '''

    lemmas = []
    lemmakeys = getlemmakeys(word)
    for lemmakey in lemmakeys:
        if lemmakey is not None:
            if lemmakey in dmldict:
                lemma = dmldict[lemmakey][1]
                lemmas.append(lemma)
            else:
                lemma = None
        else:
            lemma = None
    return lemmas


def getlemmakeys(word: str) -> List[str]:
    '''
    yields a list of lemma identifiers for the input string word
    :param word:
    :return:
    '''
    lemmakeys = []
    if word in dmwdict:
        for key in dmwdict[word]:
            featurelist = dmwkeydict[key]
            lemmakey = featurelist[3]
            lemmakeys.append(lemmakey)
    else:
        lemmakeys = []
    return lemmakeys


def getdehet(lemmakey: str) -> Optional[str]:
    '''
    yields the dehet property of the lemma identifier lemmakey
    :param lemmakey:
    :return: retrunnvalues are lexicon.de, lexicon.het, 'n/a' or None
    '''
    if lemmakey in dsldict:
        dehet = dsldict[lemmakey][dslDeHetNum]
    else:
        dehet = None
    if dehet == '':
        dehet = 'n/a'
    return dehet


def oldgetpos(lemmakey: str) -> Optional[str]:
    if lemmakey in dmldict:
        wordstructure = dmldict[lemmakey][StrucLab]
        m = posre.match(wordstructure)
        if m is not None:
            pos = m.group('pos')
        else:
            pos = 'None'
    else:
        pos = None
    return pos


def getpos(lemmakey: str) -> Optional[str]:
    '''
    returns the DCOI pt Part of Speech code for the lemma identifier lemmakey, or None
    :param lemmakey:
    :return:
    '''
    if lemmakey in dsldict:
        features = dsldict[lemmakey]
        posnum = features[dslClassNum]
        if posnum in posnum2pos:
            pos = posnum2pos[posnum]
        else:
            pos = 'None'
    return pos


def getposlist(word: str) -> List[str]:
    '''
    returns a list of pt Part of speech codes for the input string word
    :param word:
    :return:
    '''
    poslist = []
    lemmakeys = getlemmakeys(word)
    for lemmakey in lemmakeys:
        pos = getpos(lemmakey)
        poslist.append(pos)
    return poslist


def getinfls(word: str) -> List[CELEX_INFL]:
    '''
    returns a list of CELEX inflection codes for the input strin gword
    :param word:
    :return:
    '''
    infls = []
    if word in dmwdict:
        for key in dmwdict[word]:
            featurelist = dmwkeydict[key]
            infl = featurelist[4]
            infls.append(infl)
    return infls


def getwordposinfo(word: str, pos: str) -> List[WordInfo]:
    '''
    returns a list of WordInfo for the input string word with part of speech code pos
    :param word:
    :param pos:
    :return:
    '''
    cands = getwordinfo(word)
    results: List[WordInfo] = [(pt, dehet, infl, lemma) for (pt, dehet, infl, lemma) in cands if pt == pos]
    return results


def getwordinfo(word: str) -> List[WordInfo]:
    '''
    returns a lis of WordInfo for the input string word
    :param word:
    :return:
    '''
    pos_infl_lemmas: List[WordInfo] = []
    if word in dmwdict:
        for key in dmwdict[word]:
            featurelist = dmwkeydict[key]
            infl = featurelist[4]
            lemmakey = featurelist[3]
            lemma = dmldict[lemmakey][1]
            pos = getpos(lemmakey)
            dehet = getdehet(lemmakey)
            pos_infl_lemmas.append((pos, dehet, infl, lemma))
    return pos_infl_lemmas


def getinflforms(lemma: str, numClass: str, infl: str) -> List[str]:
    '''
    returns alist of inflected word forms for the input string lemma with CELEX part of speech code numClass
    and the CELEX inflectional properties infl
    :param lemma:
    :param numClass:
    :param infl:
    :return:
    '''
    results = []
    lemmakeys = dsllemmaposindex[(lemma, numClass)]
    for lemmakey in lemmakeys:
        wordkeys = dmwlemmakeyinflindex[(lemmakey, infl)]
        for wordkey in wordkeys:
            results.append(dmwkeydict[wordkey][1])
    return results



from typing import Any, Callable, Counter, Dict, List, Union

from sastadev.sastatypes import (ExactResultsDict, FileName, MatchesDict, QId,
                                 ResultsCounter, SynTree, UttWordDict)


class AllResults:
    def __init__(self, uttcount, coreresults, exactresults, postresults, allmatches, filename, analysedtrees, allutts, annotationinput=False):
        self.uttcount: int = uttcount
        self.coreresults: Dict[QId, ResultsCounter] = coreresults
        self.exactresults: ExactResultsDict = exactresults
        self.postresults: Dict[QId, Any] = postresults
        self.allmatches: MatchesDict = allmatches
        self.filename: FileName = filename
        self.analysedtrees: List[SynTree] = analysedtrees
        self.allutts: UttWordDict = allutts
        self.annotationinput: bool = annotationinput


def scores2counts(scores: Dict[QId, Counter]) -> Dict[QId, int]:
    '''
    input is a dictionary of Counter()s
    output is a dictionary of ints
    '''
    counts = {}
    for el in scores:
        countval = len(scores[el])
        counts[el] = countval
    return counts


CoreQueryFunction = Callable[[SynTree], List[SynTree]]
PostQueryFunction = Callable[[AllResults, SynTree], List[SynTree]]
QueryFunction = Union[CoreQueryFunction, PostQueryFunction]



''''
The module top3000.py creates 3 dictionaries on the basis of the input file Woordenlijsten Current.xlsx in the
subfolder *top3000*:

* semlexicon: Dict[Tuple[str, DCOIPt], List[str]] stores the semantic properties of a (lemma, pt) tuple
* trlexicon: Dict[Tuple[str, DCOIPt], List[str]] stores the transitivity properties of a (lemma, pt) tuple
* genlexicon: Dict[Tuple[str, DCOIPt], List[str]] stores whether  (lemma, pt) tuple has a genitive form (List of yes|no)

It provides functions to determine these properties for a node:

* .. autofunction:: ishuman
* .. autofunction:: isanimate
* .. autofunction:: transitive
* .. autofunction:: intransitive
* .. autofunction:: pseudotr

The top3000 lexicon file has been derived from words as listed in various resources relate to N-CDI,
in particular a top 3000 words by Schlichting. It has the following columns:

* **OrigWord** original word as it occurs in the documents
* **Word** cleaned word
* **check** check whether word is found in Origword
* **Collocate**: word or words that cooccur with the word
* **Context**: contyext as specified in the original files
* **POS**: DCOI pt or cat code
* **semtype**: semantic type, partially from the original files, so far only human, animate are used
* **Regio**: to indicated BElgium-specific words
* **trans**: transitivity: tr, intr or tr/intr, tr/intr. Alpino does not retain the valence of a verb but adpats it to the context.
* **hasgen**: has a genitive form, for nouns only: yes or no
* **Top1000**: belongs to the Top 1000 words according to Schlichting (yes, no)
* **Top3000**	N-CDI jong kk: (also) occurs in the N-CDI file for jonge kinderen (yes, no)

The information needs extension and there are more files with relevant data.
'''

import os
from typing import Dict, List, Tuple

from sastadev.conf import settings
from sastadev.namepartlexicon import namepart_isa_namepart
from sastadev.sastatypes import DCOIPt, SynTree
from sastadev.treebankfunctions import getattval
from sastadev.xlsx import getxlsxdata


def ishuman(node: SynTree) -> bool:
    '''
    The function ishuman determines whether the node node is human
    '''
    lemma = getattval(node, 'lemma')
    pt = getattval(node, 'pt')
    vwtype = getattval(node, 'vwtype')
    result = (lemma, pt) in semlexicon and 'human' in semlexicon[(lemma, pt)]
    result = result or vwtype == 'pers'
    result = result or namepart_isa_namepart(lemma)
    return result


def isanimate(node: SynTree) -> bool:
    '''
    The function isanimate determines whether the nde node is animate
    '''

    lemma = getattval(node, 'lemma')
    pt = getattval(node, 'pt')
    result = (lemma, pt) in semlexicon and 'animate' in semlexicon[(lemma, pt)]
    return result


def transitivity(node: SynTree, tr: str) -> bool:
    '''
    The function transitivity determines whether the string tr occurs in trlexicon for (lemma, pt) of the node
    '''
    lemma = getattval(node, 'lemma')
    pt = getattval(node, 'pt')
    result = (lemma, pt) in semlexicon and tr in trlexicon[(lemma, pt)]
    return result


def transitive(node: SynTree) -> bool:
    '''
    The function transitive determines whether node is transitive
    '''
    return transitivity(node, 'tr')


def pseudotr(node: SynTree) -> bool:
    '''
    The function pseudotr determines whether node is pseudotransitive
    '''
    return transitivity(node, 'tr/intr')


def intransitive(node: SynTree) -> bool:
    '''
    The function intransitve determines whether node is intransitive
    '''
    return transitivity(node, 'intr')


semicolon = ';'

filename = os.path.join(settings.SD_DIR, 'data', 'top3000', 'Woordenlijsten Current.xlsx')


lexiconheader, lexicondata = getxlsxdata(filename)

semlexicon: Dict[Tuple[str, DCOIPt], List[str]] = {}
trlexicon: Dict[Tuple[str, DCOIPt], List[str]] = {}
genlexicon: Dict[Tuple[str, DCOIPt], List[str]] = {}

for row in lexicondata:
    lemma = row[1].strip()
    pt = row[5]
    rawsems = row[6].split(semicolon)
    sems = [el.strip() for el in rawsems]
    semlexicon[(lemma, pt)] = sems

    rawtrs = row[8].split(semicolon)
    trs = [el.strip() for el in rawtrs]
    trlexicon[(lemma, pt)] = trs

    rawgens = row[9].split(semicolon)
    gens = [el.strip() for el in rawgens]
    genlexicon[(lemma, pt)] = gens

#next statement for debugging purposes
junk = 0





def sumfreq(thecounter):
    result = 0
    for (_, cnt) in thecounter.items():
        result += cnt
    return result


def getevalscores(resultscount, referencecount, intersectioncount):
    if referencecount == 0:
        recall = 100
        if resultscount == 0:
            precision = 100
        else:
            precision = 0
    else:
        recall = intersectioncount / referencecount * 100
    if resultscount == 0:
        precision = 100
        if referencecount == 0:
            recall = 100
        else:
            recall = 0
    else:
        precision = intersectioncount / resultscount * 100
    if recall + precision == 0:
        f1score = 0
    else:
        f1score = (2 * recall * precision) / (recall + precision)
    return (recall, precision, f1score)


def getscores(results, reference):
    intersection = results & reference
    lintersection = sumfreq(intersection)
    lresults = sumfreq(results)
    lreference = sumfreq(reference)
    if lreference == 0:
        recall = 100
        if lresults == 0:
            precision = 100
        else:
            precision = 0
    else:
        recall = lintersection / lreference * 100

    if lresults == 0:
        precision = 100
        if lreference == 0:
            recall = 100
        else:
            recall = 0
    else:
        precision = lintersection / lresults * 100
    if recall + precision == 0:
        f1score = 0
    else:
        f1score = (2 * recall * precision) / (recall + precision)
    return (recall, precision, f1score)



from lxml import etree

from sastadev.queryfunctions import vobij, voslashbij

# find the cases written separately


streestrings = {}
streestrings[1] = """
  <alpino_ds version="1.3">
  <metadata>
<meta type="text" name="charencoding" value="UTF8"/>
<meta type="text" name="childage" value="4;6"/>
<meta type="int" name="childmonths" value="54"/>
<meta type="text" name="comment" value="##META text title = Tarsp_01"/>
<meta type="text" name="session" value="Tarsp_01"/>
<meta type="text" name="origutt" value="allemaal varkens zit erin. "/>
<meta type="text" name="parsefile" value="Unknown_corpus_Tarsp_01_u00000000001.xml"/>
<meta type="text" name="speaker" value="CHI"/>
<meta type="int" name="uttendlineno" value="8"/>
<meta type="int" name="uttid" value="1"/>
<meta type="int" name="uttstartlineno" value="8"/>
<meta type="text" name="name" value="chi"/>
<meta type="text" name="SES" value=""/>
<meta type="text" name="age" value="4;6"/>
<meta type="text" name="custom" value=""/>
<meta type="text" name="education" value=""/>
<meta type="text" name="group" value=""/>
<meta type="text" name="language" value="nld"/>
<meta type="int" name="months" value="54"/>
<meta type="text" name="role" value="Target_Child"/>
<meta type="text" name="sex" value="male"/>
<meta type="text" name="xsid" value="1"/>
<meta type="int" name="uttno" value="1"/>
<xmeta name="tokenisation" atype="list" annotationwordlist="['allemaal', 'varkens', 'zit', 'erin', '.']" annotationposlist="[]" annotatedwordlist="[]" annotatedposlist="[]" value="['allemaal', 'varkens', 'zit', 'erin', '.']" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/><xmeta name="cleanedtokenisation" atype="list" annotationwordlist="['allemaal', 'varkens', 'zit', 'erin', '.']" annotationposlist="[]" annotatedwordlist="[]" annotatedposlist="[]" value="['allemaal', 'varkens', 'zit', 'erin', '.']" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/><xmeta name="cleanedtokenpositions" atype="list" annotationwordlist="[10, 20, 30, 40, 50]" annotationposlist="[10, 20, 30, 40, 50]" annotatedwordlist="[]" annotatedposlist="[]" value="[10, 20, 30, 40, 50]" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/></metadata>
<node begin="10" cat="top" end="51" id="0" rel="top">
    <node begin="10" cat="du" end="41" id="1" rel="--">
      <node begin="10" cat="du" end="21" id="2" rel="dp">
        <node begin="10" end="11" frame="predm_adverb" id="3" lcat="advp" lemma="allemaal" pos="adv" postag="BW()" pt="bw" rel="dp" root="allemaal" sense="allemaal" special="predm" word="allemaal"/>
        <node begin="20" end="21" frame="noun(het,count,pl)" gen="het" getal="mv" graad="basis" id="4" lcat="np" lemma="varken" ntype="soort" num="pl" pos="noun" postag="N(soort,mv,basis)" pt="n" rel="dp" rnum="pl" root="varken" sense="varken" word="varkens"/>
      </node>
      <node begin="30" cat="sv1" end="41" id="5" rel="dp">
        <node begin="30" end="31" frame="verb(hebben,sg,ld_pp)" id="6" infl="sg" lcat="sv1" lemma="zitten" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="zit" sc="ld_pp" sense="zit" stype="topic_drop" tense="present" word="zit" wvorm="pv"/>
        <node begin="40" end="41" frame="er_adverb(in)" id="7" lcat="pp" lemma="erin" pos="pp" postag="BW()" pt="bw" rel="ld" root="erin" sense="erin" special="er" word="erin"/>
      </node>
    </node>
    <node begin="50" end="51" frame="punct(punt)" id="8" lcat="--" lemma="." pos="punct" postag="LET()" pt="let" rel="--" root="." sense="." special="punt" word="."/>
  </node>
  <sentence sentid="1">allemaal varkens zit erin .</sentence></alpino_ds>

"""
streestrings[2] = """
<alpino_ds version="1.3">
  <metadata>
<meta type="text" name="charencoding" value="UTF8"/>
<meta type="text" name="childage" value="5;8"/>
<meta type="int" name="childmonths" value="68"/>
<meta type="text" name="comment" value="##META text title = Tarsp_05"/>
<meta type="text" name="session" value="Tarsp_05"/>
<meta type="text" name="origutt" value="die kan d(e)r op weer."/>
<meta type="text" name="parsefile" value="Unknown_corpus_Tarsp_05_u00000000037.xml"/>
<meta type="text" name="speaker" value="CHI"/>
<meta type="int" name="uttendlineno" value="80"/>
<meta type="int" name="uttid" value="37"/>
<meta type="int" name="uttstartlineno" value="80"/>
<meta type="text" name="name" value="chi"/>
<meta type="text" name="SES" value=""/>
<meta type="text" name="age" value="5;8"/>
<meta type="text" name="custom" value=""/>
<meta type="text" name="education" value=""/>
<meta type="text" name="group" value=""/>
<meta type="text" name="language" value="nld"/>
<meta type="int" name="months" value="68"/>
<meta type="text" name="role" value="Target_Child"/>
<meta type="text" name="sex" value="female"/>
<meta type="text" name="xsid" value="37"/>
<meta type="int" name="uttno" value="37"/>
<xmeta name="Noncompletion of a Word" atype="text" annotationwordlist="['(e)']" annotationposlist="[]" annotatedwordlist="[]" annotatedposlist="[]" value="['(e)']" cat="None" subcat="None" source="None" backplacement="0" penalty="10"/><xmeta name="tokenisation" atype="list" annotationwordlist="['die', 'kan', 'd(e)r', 'op', 'weer', '.']" annotationposlist="[]" annotatedwordlist="[]" annotatedposlist="[]" value="['die', 'kan', 'd(e)r', 'op', 'weer', '.']" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/><xmeta name="cleanedtokenisation" atype="list" annotationwordlist="['die', 'kan', 'der', 'op', 'weer', '.']" annotationposlist="[]" annotatedwordlist="[]" annotatedposlist="[]" value="['die', 'kan', 'der', 'op', 'weer', '.']" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/><xmeta name="cleanedtokenpositions" atype="list" annotationwordlist="[10, 20, 30, 40, 50, 60]" annotationposlist="[10, 20, 30, 40, 50, 60]" annotatedwordlist="[]" annotatedposlist="[]" value="[10, 20, 30, 40, 50, 60]" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/></metadata>
<node begin="10" cat="top" end="61" id="0" rel="top">
    <node begin="10" cat="smain" end="51" id="1" rel="--">
      <node begin="10" end="11" frame="determiner(de,nwh,nmod,pro,nparg)" getal="getal" id="2" infl="de" lcat="np" lemma="die" naamval="stan" pdtype="pron" persoon="3" pos="det" postag="VNW(aanw,pron,stan,vol,3,getal)" pt="vnw" rel="su" rnum="sg" root="die" sense="die" status="vol" vwtype="aanw" wh="nwh" word="die"/>
      <node begin="20" end="21" frame="verb(hebben,modal_not_u,ld_pp)" id="3" infl="modal_not_u" lcat="smain" lemma="kunnen" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="kan" sc="ld_pp" sense="kan" stype="declarative" tense="present" word="kan" wvorm="pv"/>
      <node begin="30" cat="pp" end="41" id="4" rel="ld">
        <node begin="30" end="31" frame="er_vp_adverb" getal="getal" id="5" lcat="advp" lemma="er" naamval="stan" pdtype="adv-pron" persoon="3" pos="adv" postag="VNW(aanw,adv-pron,stan,red,3,getal)" pt="vnw" rel="obj1" root="er" sense="er" special="er" status="red" vwtype="aanw" word="der"/>
        <node begin="40" end="41" frame="preposition(op,[af,na])" id="6" lcat="pp" lemma="op" pos="prep" postag="VZ(fin)" pt="vz" rel="hd" root="op" sense="op" vztype="fin" word="op"/>
      </node>
      <node begin="50" end="51" frame="adverb" id="7" lcat="advp" lemma="weer" pos="adv" postag="BW()" pt="bw" rel="mod" root="weer" sense="weer" word="weer"/>
    </node>
    <node begin="60" end="61" frame="punct(punt)" id="8" lcat="punct" lemma="." pos="punct" postag="LET()" pt="let" rel="--" root="." sense="." special="punt" word="."/>
  </node>
  <sentence sentid="37">die kan der op weer .</sentence></alpino_ds>
"""
streestrings[3] = """
  <alpino_ds version="1.3">
  <metadata>
<meta type="text" name="charencoding" value="UTF8"/>
<meta type="text" name="childage" value="5;2"/>
<meta type="int" name="childmonths" value="62"/>
<meta type="text" name="comment" value="##META text title = Tarsp_04"/>
<meta type="text" name="session" value="Tarsp_04"/>
<meta type="text" name="origutt" value="die [: daar] heb wij ook een boekje van"/>
<meta type="text" name="parsefile" value="Unknown_corpus_Tarsp_04_u00000000002.xml"/>
<meta type="text" name="speaker" value="CHI"/>
<meta type="int" name="uttendlineno" value="10"/>
<meta type="int" name="uttid" value="2"/>
<meta type="int" name="uttstartlineno" value="10"/>
<meta type="text" name="name" value="chi"/>
<meta type="text" name="SES" value=""/>
<meta type="text" name="age" value="5;2"/>
<meta type="text" name="custom" value=""/>
<meta type="text" name="education" value=""/>
<meta type="text" name="group" value=""/>
<meta type="text" name="language" value="nld"/>
<meta type="int" name="months" value="62"/>
<meta type="text" name="role" value="Target_Child"/>
<meta type="text" name="sex" value="female"/>
<meta type="text" name="xsid" value="2"/>
<meta type="int" name="uttno" value="2"/>
<xmeta name="GrammarError" atype="text" annotationwordlist="['hebben']" annotationposlist="[50]" annotatedwordlist="['heb']" annotatedposlist="[50]" value="SVAerror" cat="Error" subcat="None" source="SASTA" backplacement="2" penalty="10"/><xmeta name="Replacement" atype="text" annotationwordlist="['daar']" annotationposlist="[30]" annotatedwordlist="['die']" annotatedposlist="[10]" value="['daar']" cat="None" subcat="None" source="CHAT" backplacement="0" penalty="10"/><xmeta name="tokenisation" atype="list" annotationwordlist="['die', '[: ', 'daar', ']', 'heb', 'wij', 'ook', 'een', 'boekje', 'van']" annotationposlist="[]" annotatedwordlist="[]" annotatedposlist="[]" value="['die', '[: ', 'daar', ']', 'heb', 'wij', 'ook', 'een', 'boekje', 'van']" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/><xmeta name="cleanedtokenisation" atype="list" annotationwordlist="['daar', 'heb', 'wij', 'ook', 'een', 'boekje', 'van']" annotationposlist="[]" annotatedwordlist="[]" annotatedposlist="[]" value="['daar', 'heb', 'wij', 'ook', 'een', 'boekje', 'van']" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/><xmeta name="cleanedtokenpositions" atype="list" annotationwordlist="[30, 50, 60, 70, 80, 90, 100]" annotationposlist="[30, 50, 60, 70, 80, 90, 100]" annotatedwordlist="[]" annotatedposlist="[]" value="[30, 50, 60, 70, 80, 90, 100]" cat="None" subcat="None" source="CHAT/Tokenisation" backplacement="0" penalty="10"/><xmeta name="parsed_as" atype="text" annotationwordlist="daar hebben wij ook een boekje van" annotationposlist="[]" annotatedwordlist="[]" annotatedposlist="[]" value="daar hebben wij ook een boekje van" cat="Correction" subcat="None" source="SASTA" backplacement="0" penalty="0"/></metadata>
<node begin="30" cat="top" end="101" id="0" rel="top">
    <node begin="30" cat="smain" end="101" id="1" rel="--">
      <node begin="50" end="51" frame="verb(hebben,sg1,transitive_ndev)" id="1" infl="sg1" lcat="--" lemma="hebben" pos="verb" postag="WW(pv,tgw,ev)" pt="ww" pvagr="ev" pvtijd="tgw" rel="hd" root="heb" sc="transitive_ndev" sense="heb" tense="present" word="heb" wvorm="pv"/>
    <node begin="60" case="nom" def="def" end="61" frame="pronoun(nwh,fir,pl,de,nom,def)" gen="de" getal="mv" id="3" lcat="np" lemma="wij" naamval="nomin" num="pl" pdtype="pron" per="fir" persoon="1" pos="pron" postag="VNW(pers,pron,nomin,vol,1,mv)" pt="vnw" rel="su" rnum="pl" root="wij" sense="wij" status="vol" vwtype="pers" wh="nwh" word="wij"/>
      <node begin="70" end="71" frame="sentence_adverb" id="4" lcat="advp" lemma="ook" pos="adv" postag="BW()" pt="bw" rel="mod" root="ook" sense="ook" special="sentence" word="ook"/>
      <node begin="80" cat="np" end="91" id="5" rel="obj1">
        <node begin="80" end="81" frame="determiner(een)" id="6" infl="een" lcat="detp" lemma="een" lwtype="onbep" naamval="stan" npagr="agr" pos="det" postag="LID(onbep,stan,agr)" pt="lid" rel="det" root="een" sense="een" word="een"/>
        <node begin="90" end="91" frame="noun(het,count,sg)" gen="het" genus="onz" getal="ev" graad="dim" id="7" lcat="np" lemma="boek" naamval="stan" ntype="soort" num="sg" pos="noun" postag="N(soort,ev,dim,onz,stan)" pt="n" rel="hd" rnum="sg" root="boek_DIM" sense="boek_DIM" word="boekje"/>
      </node>
      <node begin="30" cat="pp" end="101" id="8" rel="pc">
        <node begin="30" end="31" frame="er_loc_adverb" getal="getal" id="9" lcat="advp" lemma="daar" naamval="obl" pdtype="adv-pron" persoon="3o" pos="adv" postag="VNW(aanw,adv-pron,obl,vol,3o,getal)" pt="vnw" rel="obj1" root="daar" sense="daar" special="er_loc" status="vol" vwtype="aanw" word="daar"/>
        <node begin="100" end="101" frame="preposition(van,[af,uit,vandaan,[af,aan]])" id="10" lcat="pp" lemma="van" pos="prep" postag="VZ(fin)" pt="vz" rel="hd" root="van" sense="van" vztype="fin" word="van"/>
      </node>
    </node>
  </node>
  <sentence sentid="2">daar heb wij ook een boekje van</sentence><comments>
    <comment>Q#ng1664379677|daar hebben wij ook een boekje van|1|1|-3.187513080109999</comment>
  </comments>
</alpino_ds>

"""

strees = {i: etree.fromstring(ts) for i, ts in streestrings.items()}

# reference dictionary contains tuples with (1) VoBij reference results; (2) Voslashbij reference results
reference = {}
reference[1] = (['40'], [])
reference[2] = (['30'], [])
reference[3] = ([], ['30'])


def test():
    for i, stree in strees.items():
        results = [n.attrib['begin'] for n in vobij(stree)]
        try:
            assert results == reference[i][0]
        except AssertionError as e:
            print(
                f'Vobij: i={i}, reference={reference[i][0]}, =/= results={results}')
            raise AssertionError

        results = [n.attrib['begin'] for n in voslashbij(stree)]
        try:
            assert results == reference[i][1]
        except AssertionError as e:
            print(
                f'Vo/bij: i={i}, reference={reference[i][1]}, =/= results={results}')
            raise AssertionError


def main():
    # define tests
    test()


if __name__ == '__main__':
    main()



import os
import re

from sastadev.xlsx import getxlsxdata


def remove(patlist, rawstr):
    result = rawstr
    for pat in patlist:
        result = re.sub(pat, '', result)
    return result


def clean(rawutt):
    pats = [r'dd:', r'\[', r'\]', r'<', r'>', r'0', r'\*', r'c:', r'\+', r'com:', r'=']
    result = remove(pats, rawutt)
    result = re.sub(r'(?i)iemand\s*\|\s*iets', 'iemand', result)
    result = re.sub(r'(?i)iets\s*\|\s*iemand', 'iemand', result)
    result = re.sub(r'\u2018', "'", result)
    result = re.sub(r'\u2019', "'", result)

    return result


metamodel = '##META text {att} = {val}'

inpath = r'C:\Users\Odijk101\Dropbox\jodijk\Utrecht\researchproposals\MWEs'
infilename = 'UitdrukkingenNL expanded_dedup Current.xlsx'

infullname = os.path.join(inpath, infilename)

rawheader, data = getxlsxdata(infullname, sheetname='Data')

header = [re.sub(r'\s', '_', rh) for rh in rawheader]

utterancecol = 4  # kolom E

peplines = []
for row in data:
    lrow = len(row)
    for i in range(lrow):
        att = header[i]
        val = row[i]
        newmeta = metamodel.format(att=att, val=val)
        peplines.append(newmeta)
    rawutt = row[utterancecol]
    cleanutt = clean(rawutt)
    peplines.append(cleanutt)
    peplines.append('')     # blank line to finish input for this utterance

outpath = inpath
outfilename = 'MWE2022-08-03.txt'
outfullname = os.path.join(outpath, outfilename)
with open(outfullname, 'w', encoding='utf8') as outfile:
    for pepline in peplines:
        print(pepline, file=outfile)



from collections import Counter, defaultdict
from math import isnan

from sastadev import readcsv
from sastadev.conf import settings
from sastadev.counterfunctions import counter2liststr
from sastadev.xlsx import getxlsxdata, mkworkbook

permprefix = 'perm_'

permsilvercolcount = 11   # number of columns in the silverperm files

user1col = 0
user2col = 1
user3col = 2
moreorlesscol = 3
qidcol = 4
uttidcol = 8
poscol = 9

uttidscol = 4  # in paltinum-edited tsv files

nots = ['not']
oks = ['ok', 'oke']
undecideds = ['?', 'ok/not', 'not/ok']
allowedoknots = oks + nots + undecideds
legalmoreorlesses = ['More examples', 'Missed examples']
comma = ','
commaspace = ', '

platinumchecksuffix = '_platinum.check.tsv'
platinumcheckeditedsuffix = '_platinum.check-edited.tsv'
platinumsuffix = '.platinum.tsv'
platinumeditedsuffix = '.platinum-edited.tsv'


def write2excel(datadict, header, filename):
    data = [datadict[key] for key in datadict]
    workbook = mkworkbook(filename, [header], data)
    workbook.close()


def getheader(data):
    if data is None:
        result = []
    else:
        result = data.head()
    return result


def checkpermformat(header, data, colcount, strict=True):
    result = True
    lheader = len(header)
    if (lheader == 0 or header == ['']) and data == []:
        return True
    result = lheader == colcount
    if result:
        rowctr = 0
        for row in data:
            rowctr += 1
            lrow = len(row)
            result = lrow == colcount
            if not result:
                settings.LOGGER.error('Wrong # columns ({} instead of {}), row {}'.format(lrow, colcount, rowctr))
                if strict:
                    exit(-1)
                else:
                    return False
    else:
        settings.LOGGER.error('Wrong # columns ({} instead of {}) in the header'.format(lheader, colcount,))
        if strict:
            exit(-1)
        else:
            return False
    return result


def updatepermdict(fullname, permdict):
    silverheader, silverdata = getxlsxdata(fullname)
    colsok = checkpermformat(silverheader, silverdata, permsilvercolcount, strict=False)
    silverfulldatadict = silverdata2dict(silverdata)

    #Voeg silverfulldatadict toe aan permdict
    for key in silverfulldatadict:
        if key not in permdict:
            permdict[key] = silverfulldatadict[key]
        elif not rowsequal(silverfulldatadict[key], permdict[key]):
            settings.LOGGER.warning('Key: {} Value:\n ({}) \noverwritten by value:\n {};\n File: {}'.format(key, permdict[key], silverfulldatadict[key], fullname))

    return permdict, silverheader


def rowsequal(row1, row2, casesensitive=False):
    if len(row1) != len(row2):
        return False
    pairs = zip(row1, row2)
    for el1, el2 in pairs:
        if isinstance(el1, str) and isinstance(el2, str):
            if el1.lower() != el2.lower():
                return False
        elif el1 != el2:
            return False
    return True


def getsilverannotations(perm_silverfullname, platinumcheckeditedfullname,
                         platinumcheckfullname, silvercheckfullname,
                         platinumfullname, platinumeditedfullname, goldscores):

    # -lees perm_silverfulldata in, voeg toe aan perm_silverfulldatadict
    perm_silverfulldatadict = dict()
    perm_silverfulldatadict, perm_header = updatepermdict(perm_silverfullname, perm_silverfulldatadict)

    # lees platinumcheckeditedfilename in, voeg toe aan perm_silverfulldatadict
    perm_silverfulldatadict, silverheader = updatepermdict(platinumcheckeditedfullname, perm_silverfulldatadict)

    #-lees platinmumcheck in, voeg toe aan perm_silverfulldatadict
    perm_silverfulldatadict, silverheader = updatepermdict(platinumcheckfullname, perm_silverfulldatadict)

    #-schrijf de prem_silverfulldatadict weg naar een nieuwe perm_silverfilename
    write2excel(perm_silverfulldatadict, silverheader, perm_silverfullname)

    #maak de silver file (platinum)--zie functie mksilver maar herschreven want je hebt de files al gelezen
    #@@
    mksilver(perm_silverfulldatadict, silvercheckfullname, platinumfullname, platinumeditedfullname, goldscores)

    #loop alle entries af in  platinumcheck, indien in perm_silverfulldatadict, [User 1-3] + entry{3:]
    # print dit naar de platinumcheckfile@@
    # of return de perm_silverfulldatadict hier en doe bovenstaande in de main loop waar je de file nu al uitschrijft
    return perm_silverfulldatadict


def clean(inval):
    #if type(inval) != str:
    #    print('nonstring value: {}'.format(inval))
    instr = str(inval)
    result = instr.strip()
    result = result.lower()
    return result


def listminus(list1, list2):
    clist1 = Counter(list1)
    clist2 = Counter(list2)
    cresult = clist1 - clist2
    result = counter2liststr(cresult)
    return result


def nono(inval):
    result = (inval is None) or (inval == 0) or (inval == []) or (inval == '')
    return result


def myisnan(inval):
    try:
        result = isnan(inval)
    except Exception:
        result = False
    return result


def silverdata2dict(silverdata):
    #make a dictionary out of data: a list of rows
    #silverdict = dict()
    silverfulldatadict = dict()
    if silverdata is not None:
        maxrow = len(silverdata)
        for rowctr in range(maxrow):
            therow = silverdata[rowctr]
            user1 = therow[user1col]
            user2 = therow[user2col]
            user3 = therow[user3col]
            qid = therow[qidcol]
            uttid = str(therow[uttidcol])
            pos = therow[poscol]
            thekey = (qid, uttid, pos)
            # only add it when any of user1, user2, user3 has a nonempty value
            if not (nono(user1) and nono(user2) and nono(user3)):
                #silverdict[thekey] = (user1, user2, user3)
                silverfulldatadict[thekey] = silverdata[rowctr]
    return silverfulldatadict  # , silverdict


def mksilver(permsilverdict, silvercheckfullname, platinumfullname, platinumeditedfullname, goldscores):

    # read the silvercheckfile
    silverheader, silvercheckdata = getxlsxdata(silvercheckfullname)

    # determine which uttids have to be removed
    undecidedcounter = 0
    toremove = defaultdict(list)
    maxrow = len(silvercheckdata)
    for row in range(maxrow):
        currow = silvercheckdata[row]
        moreorless = currow[moreorlesscol]
        if moreorless not in legalmoreorlesses:
            settings.LOGGER.error('Unexpected value in row {}: {}. File {}'.format(row, moreorless, silvercheckfullname))
        if moreorless == 'Missed examples':
            continue
        qid = currow[qidcol]
        uttid = str(currow[uttidcol])
        pos = currow[poscol]
        if (qid, uttid, pos) in permsilverdict:
            curpermrow = permsilverdict[(qid, uttid, pos)]
            (user1, user2, user3) = curpermrow[user1col], curpermrow[user2col], curpermrow[user3col]
            cleanuser1 = clean(user1)
            if cleanuser1 not in allowedoknots:
                settings.LOGGER.error('Unexpected value in row {}: {}. File {}'.format(row, user1, silvercheckfullname))
            if cleanuser1 not in oks:
                toremove[qid].append(uttid)
            if cleanuser1 in undecideds:
                undecidedcounter += 1
        else:
            pass
            #settings.LOGGER.warning('No Silver remark for row {}: {}. File {}; qid={}, uttid={}, pos={}'.format(row, moreorless, silvercheckfullname, qid, uttid, pos))
    junk = 0
    if undecidedcounter > 0:
        settings.LOGGER.info('{} undecided in file {}'.format(undecidedcounter, silvercheckfullname))
    # read the platinumfile
    (header, platinumdata) = readcsv.readheadedcsv(platinumfullname)
    newrows = []
    for (rowctr, row) in platinumdata:
        theqid = row[0]
        if theqid in toremove:
            toremoveids = [str(x) for x in toremove[theqid]]
            olduttids_string = row[uttidscol]
            rawolduttids = olduttids_string.split(comma)
            olduttids = [clean(x) for x in rawolduttids]
            toremoveCounter = Counter(toremoveids)
            olduttsCounter = Counter(olduttids)
            goldcounter = goldscores[theqid][2] if theqid in goldscores else Counter()
            newCounter = (olduttsCounter - toremoveCounter) | goldcounter  # all gold results must stay in
            newuttids_string = counter2liststr(newCounter)
            #newuttids_string = listminus(olduttids, toremoveids)
            newrow = row[:4] + [newuttids_string] + row[uttidscol + 1:]
        else:
            newrow = row
        newrows.append(newrow)

    # write the results to a new edited platinumfile

    readcsv.writecsv(newrows, platinumeditedfullname, header=header)



commaspace = ', '


def checkinteger(thestr):
    try:
        cast = int(thestr)
        result = True
    except ValueError:
        result = False
    return result


def counter2list(thecounter):
    resultlist = []
    allintegers = True
    for el in thecounter:
        if allintegers:
            allintegers = allintegers and checkinteger(el)
        cnt = thecounter[el]
        for i in range(cnt):
            resultlist.append(str(el))
    if allintegers:
        sortedresultlist = sorted(resultlist, key=int)
    else:
        sortedresultlist = sorted(resultlist)
    return sortedresultlist


def counter2liststr(thecounter):
    thelist = counter2list(thecounter)
    result = commaspace.join(thelist)
    return result



import os
import re
from collections import Counter
from io import BytesIO

import xlsxwriter

from sastadev import xlsx
from sastadev.conf import settings
from sastadev.counterfunctions import counter2liststr
from sastadev.forms import getformfilename

ordA = ord('A')
comma = ','
xlsxext = '.xlsx'

idpat = r'^[TSA][0-9]{3}$'
idcpat = r'^[TSA][0-9]{3}c$'
idre = re.compile(idpat)
idcre = re.compile(idcpat)

tarspformsuffix = '_TARSP-Form'

#tarspformsuffixext = tarspformsuffix + xlsxext
#intreebanksfolder = 'intreebanks'


def getshortloc(colctr, rowctr):
    #colctr must be smaller than 26
    colstr = chr((colctr % 26) + ordA)
    rowstr = str(rowctr + 1)
    result = colstr + rowstr
    return result


# def oldreadbaseform(infilename):
#     basesheet = {}
#     wb = xlrd.open_workbook(infilename)
#     sheet = wb.sheet_by_index(0)
#     startrow = 0
#     startcol = 0
#     lastrow = sheet.nrows
#     lastcol = sheet.ncols
#     for rowctr in range(startrow, lastrow):
#         for colctr in range(startcol, lastcol):
#             curval = sheet.cell_value(rowctr, colctr)
#             if curval is not None and curval != '':
#                 basesheet[(rowctr, colctr)] = curval
#     return basesheet

def readbaseform(infilename):
    basesheet = {}
    header, data = xlsx.getxlsxdata(infilename)
    for rowctr, row in enumerate(data):
        for colctr, curval in enumerate(row):
            if curval is not None and curval != '':
                basesheet[(rowctr, colctr)] = curval
    return basesheet


def is_id(word):
    result = idre.match(word)
    return result


def is_idc(word):
    result = idcre.match(word)
    return result


def idc2id(word):
    if word[-1] == 'c':
        result = word[:-1]
    else:
        result = word
    return result


def getval(allresults, idx):
    if idx in allresults.coreresults:
        result = allresults.coreresults[idx]
    elif idx in allresults.postresults:
        result = allresults.postresults[idx]
    else:
        result = ''
    return result


def val2str(aval):
    if isinstance(aval, Counter):
        result = counter2liststr(aval)
    elif isinstance(aval, list):
        result = comma.join(aval)
    else:
        result = str(aval)
    return result


def mktarspform(allresults, _, in_memory=False):
    global basesheet

    if not in_memory:
        target = getformfilename(allresults.filename, tarspformsuffix)
        #(base, ext) = os.path.splitext(allresults.filename)
        #core, filename = os.path.split(base)
        #root, lastfolder = os.path.split(core)
        #if lastfolder == intreebanksfolder:
        #    target = os.path.join(root, 'forms', filename + tarspformsuffixext)
        #else:
        #   target = base + tarspformsuffixext
    else:
        target = BytesIO()

    workbook = xlsxwriter.Workbook(target, {"strings_to_numbers": True})
    worksheet = workbook.add_worksheet()

    for (rowctr, colctr) in basesheet:
        curval = str(basesheet[(rowctr, colctr)])
        if is_id(curval):
            newval = getval(allresults, curval)
            # write newval to the new sheet
            newvalstr = val2str(newval)
            worksheet.write(rowctr, colctr, newvalstr)
        elif is_idc(curval):
            urval = idc2id(curval)
            newval = getval(allresults, urval)
            cval = len(newval)
            newvalstr = val2str(cval)
            worksheet.write(rowctr, colctr, newvalstr)
        else:
            worksheet.write(rowctr, colctr, curval)

    #formatting
    textwrap = workbook.add_format()
    textwrap.set_text_wrap()
    boldbottom = workbook.add_format()
    boldbottom.set_bottom(5)
    textwrapcolumns = [2, 5, 8, 11, 14, 17, 20, 23]  # CFILORUX
    for col in textwrapcolumns:
        worksheet.set_column(col, col, None, textwrap)
    boldbottomrows = [12, 20, 28, 39, 44, 53, 56]
    for row in boldbottomrows:
        worksheet.set_row(row, row, boldbottom)

    columnwidths = {'B': 12, 'C': 14, 'D': 3, 'G': 3, 'H': 12, 'J': 3, 'K': 9, 'M': 3, 'N': 12, 'P': 3, 'S': 3, 'V': 3, 'W': 12, 'Y': 3, 'Z': 3}
    for row in columnwidths:
        range = row + ':' + row
        worksheet.set_column(range, columnwidths[row])

    # worksheet.set_landscape()
    worksheet.fit_to_pages(1, 1)
    worksheet.hide_gridlines(0)  # do not hide gridlines

    workbook.close()
    return target


# initialisation
basefilename = os.path.join(settings.SD_DIR, 'data', 'form_templates', 'TARSP Form Current.xlsx')
basesheet = readbaseform(basefilename)



'''
The lexicon module is the interface to the lexicon.
It is intended to abstract from the concrete lexicon used.

Currently we especially use the CELEX lexicon.
This module also contains some special word lists. Perhaps we should set up a special Exception List module
for this purpose.


'''

from typing import Any, Dict, List, Optional

from sastadev import celexlexicon, treebankfunctions
from sastadev.namepartlexicon import (namepart_isa_namepart,
                                      namepart_isa_namepart_uc)
from sastadev.sastatypes import CELEX_INFL, DCOITuple, Lemma, SynTree, WordInfo

space = ' '

celex = 'celex'
alpino = 'alpino'

# the CHAT codes xxx and yyy must be recognised as valid codes and as valid words in some cases
chatspecials = ['xxx', 'yyy']


lexicon = celex

#Alpino often analyses certain words as tsw though they should be analysed as nouns
tswnouns = ['baby', 'jongen', 'juf', 'jufforouw', 'mam', 'mama', 'mamma', 'meisje', 'mens', 'meneer', 'mevrouw',
            'pap', 'papa', 'pappa', 'stouterd', 'opa', 'oma']

de = '1'
het = '2'
# List of de-determiners, List of corresponding het-determiners, and implicitly by this a mapping between the two
dets = {}
dets[de] = ['de', 'die', 'deze', 'onze', 'welke', 'iedere', 'elke', 'zulke']
dets[het] = ['het', 'dat', 'dit', 'ons', 'welk', 'ieder', 'elk', 'zulk']


def isa_namepart(word: str) -> bool:
    '''
    is the word a name part
    :param word:
    :return:
    '''
    return namepart_isa_namepart(word)


def isa_namepart_uc(word: str) -> bool:
    '''
    is the word in upper case a name part
    :param word:
    :return:
    '''
    return namepart_isa_namepart_uc(word)


def lookup(dct: Dict[str, Any], key: str) -> str:
    '''
    looks up key in dct, if so it returns dct[key] else ''
    :param dct:
    :param key:
    :return:
    '''
    result = dct[key] if key in dct else ''
    return result


def pvinfl2dcoi(word: str, infl: CELEX_INFL, lemma: Lemma) -> Optional[DCOITuple]:
    '''
    encodes the CELEX code infl for word (which must be a pv) as a DCOI Tuple
    at least if the CELEX lexicon is used, else None
    :param word:
    :param infl:
    :param lemma:
    :return:
    '''
    if lexicon == celex:
        results = celexlexicon.celexpv2dcoi(word, infl, lemma)
        wvorm = lookup(results, 'wvorm')
        pvtijd = lookup(results, 'pvtijd')
        pvagr = lookup(results, 'pvagr')
        positie = lookup(results, 'positie')
        buiging = lookup(results, 'buiging')
        dcoi_infl = []
        atts = [wvorm, pvtijd, pvagr, positie, buiging]
        for att in atts:
            if att != '':
                dcoi_infl.append(att)
        result = tuple(dcoi_infl)
    else:
        result = None
    return result


def getwordposinfo(word: str, pos: str) -> List[WordInfo]:
    '''
    yields the list of WordInfo for word str with part of speech code pos by looking it up in the lexicon in use
    :param word:
    :param pos:
    :return:
    '''
    results = []
    if lexicon == celex:
        results = celexlexicon.getwordposinfo(word, pos)
    return results


def getwordinfo(word: str) -> List[WordInfo]:
    '''
    yields the list of WordInfo for word str  by looking it up in the lexicon in use
    :param word:
    :return:
    '''
    results = []
    if lexicon == celex:
        results = celexlexicon.getwordinfo(word)
    return results


def informlexicon(word: str) -> bool:
    '''
    checks whether word is in the  word form lexicon
    :param word:
    :return:
    '''
    allwords = word.split(space)
    result = True
    for aword in allwords:
        if lexicon == 'celex':
            result = result and celexlexicon.incelexdmw(aword)
        elif lexicon == 'alpino':
            result = False
        else:
            result = False
    return result


def informlexiconpos(word: str, pos: str) -> bool:
    '''
    checks whether word with part of speech code pos is in the word form lexicon
    :param word:
    :param pos:
    :return:
    '''

    allwords = word.split(space)
    result = True
    for aword in allwords:
        if lexicon == 'celex':
            result = result and celexlexicon.incelexdmwpos(aword, pos)
        elif lexicon == 'alpino':
            result = False
        else:
            result = False
    return result


def chatspecial(word: str) -> bool:
    result = word in chatspecials
    return result


def known_word(word: str) -> bool:
    '''
    a word is considered to be a known_word if it occurs in the word form lexicon,
    if it is a name part, or if it is a chatspecial item
    :param word:
    :return:
    '''
    result = informlexicon(word) or isa_namepart(word) or chatspecial(word)
    return result


def getinflforms(thesubj: SynTree, thepv: SynTree, inversion: bool) -> List[str]:
    '''
    yields the list of  finite verb word forms that
    -agrees with the subject node (thesubj),
    -has the same lemma as the word form in the pv node
    -is compatible with whether there is inversion or not
    :param thesubj:
    :param thepv:
    :param inversion:
    :return:
    '''
    if lexicon == 'celex':
        pt = treebankfunctions.getattval(thepv, 'pt')
        pos = celexlexicon.pos2posnum[pt]
        infl = celexlexicon.dcoiphi2celexpv(thesubj, thepv, inversion)
        lemma = treebankfunctions.getattval(thepv, 'lemma')
        results = celexlexicon.getinflforms(lemma, pos, infl)
    else:
        results = []
    return results



'''
The module *Sziplus* implements two TARSP language measures:

* Vr5+: through the function *vr5plus*:

  .. autofunction:: Sziplus::vr5plus

  In the meantime a different implementation using macros has replaced this function,
  so it has become obsolete

* 6+: through the function *sziplus6*:

  .. autofunction:: Sziplus::sziplus6

'''


from typing import List

from sastadev.sastatypes import SynTree
from sastadev.treebankfunctions import getattval

#: the variable (constant) *clausequery*  is an XPath query that searches for nodes
#: that are clauses.
clausequery = './/node[@cat="smain" or @cat="ssub" or @cat="sv1" or ' \
              '@cat="whq" or @cat="whrel" or @cat="whsub" or @cat="cp" ]'

#: the variable (constant) *vrquery* is an XPath query that searchs for nodes that are
#: wh-questions.
vrquery = './/node[@cat="whq" or @cat="whsub" or (@cat="whrel" and ../node[@cat="top"])]'


def empty(alist: list) -> bool:
    return (alist == [])


def notempty(alist: list) -> bool:
    return (not empty(alist))


def noposcatin(node: SynTree) -> bool:
    '''
    The function *noposcatin* checks whether a node has none of the following
    attributes: *pt*, *pos*, *cat*
    '''
    result = 'pt' not in node.attrib and 'pos' not in node.attrib and 'cat' not in node.attrib
    return result


def isindexnode(node: SynTree) -> bool:
    '''
    The function *isinexnode* determines whether a node is an *index node*. It is if:

    * the node has an *index* attribute
    * the node has no part of speech or syntactic category attribute (as determined by the function *noposcatin*).

    The function *noposcatin* is defined as follows:

    .. autofunction:: Sziplus::noposcatin

    * **Remark** The function *noposcatin* is better replaced by a function that checks for the absence of the attributes *cat* and *word*.

    '''
    result = 'index' in node.attrib and noposcatin(node)
    return result


def isvcinforppart(node: SynTree) -> bool:
    '''
    The function *isvcinforppart* determines whether a node is a node for a nonfinite
    verbal complement. That is the case if

    *  its category is one of *inf*, *teinf*, or *ppart*
    * its relation has the value *vc*

    '''
    rel = getattval(node, 'rel')
    cat = getattval(node, 'cat')
    if rel == 'vc' and cat in ['inf', 'teinf', 'ppart']:
        result = True
    else:
        result = False
    return result


def isrealnode(node: SynTree) -> bool:
    '''
    The fucntion *isrealnode* determines whether a nide is a real node, which it is if:

    * it is not a node for an interpunction sign
    * it is not a nonfinite complement
    * if it is not a separable particle word of a verb
    * if it is not an index node 9as detemined by the function *isindexnode*)

    The function *isindexnode* is defined as follows:

    .. autofunction:: Sziplus::isindexnode
    '''
    pt = getattval(node, 'pt')
    rel = getattval(node, 'rel')
    if pt == 'let':
        result = False
    elif isvcinforppart(node):
        result = False
    elif rel == 'svp' and pt in node.attrib:
        result = False
    elif isindexnode(node):
        result = False
    else:
        result = True
    return result


def isbodysv1(node: SynTree) -> bool:
    if node is None:
        result = False
    else:
        result = 'cat' in node.attrib and node.attrib['cat'] in ['sv1', 'ssub'] \
                 and 'rel' in node.attrib and node.attrib['rel'] == 'body'
    return result


def getnodecount(clause: SynTree) -> int:
    '''
    The function *getnodecount* counts the number of real nodes in *clause*. It
    considers each child of *clause*:

    * if the child is a real node (as determined by the function *isrealnode*, the counter goes up by 1

     * if it is an *sv1* or *ssub* clause body, the function *getnodecount* is recursively applied to the child, and the result added to the counter. This is for structures such whq[ *wie* body/sv1[ *heeft dat gedaan* ]] and cp[ *omdat* body/ssub[ *hij ziek is* ]]

    * if it is a nonfinite complement (as determined by the function *isvcinforppart*), the function *getnodecount* is recursively applied to child, the result added to the counter, minus 1 (we do not count the verb inside a nonfinite complement because it is part of the "gezegde".


    The function *isrealnode* is defined as follows:

    .. autofunction:: Sziplus::isrealnode

    The function *isvcinforppart* is defined as follows:

    .. autofunction:: Sziplus::isvcinforppart

    '''
    nodectr = 0
    for child in clause:
        if isrealnode(child):
            nodectr += 1
        elif isbodysv1(child):
            nodectr += getnodecount(child)
        elif isvcinforppart(child):
            nodectr += getnodecount(child) - 1  # we do no count the verb because it is part of the 'gezegde'
    return nodectr


def sziplus(syntree: SynTree, i: int) -> List[SynTree]:
    '''
    The function *sziplus* takes a SynTree *syntree* and an integer *i* and uses the
    function *nodeiplus* by applying it to  *syntree*, *i*, and *clausequery*:

    .. autodata:: Sziplus::clausequery

    The function *nodeiplus* is defined as follows:

    .. autofunction:: Sziplus::nodeiplus

    '''
    results = nodeiplus(syntree, i, clausequery)
    return results

# vr5plus is obsolete because there is composed language measure for it.


def vr5plus(syntree: SynTree) -> List[SynTree]:
    '''
    The function *vr5plus* is intended to implement the TARSP language measure *Vr5+*.
    it does so by applying the function *nodeiplus* to the syntree, integr 5 and the
    Xpath query *vrquery*.

    The Xpath *vrquery* is defined as follows:

    .. autodata:: Sziplus::vrquery

    The function *nodeiplus* is defined as follows:

    .. autofunction:: Sziplus::nodeiplus
    '''
    results = nodeiplus(syntree, 5, vrquery)
    return results


def nodeiplus(syntree: SynTree,
              i: int,
              query: str) -> List[SynTree]:
    '''
    The function *nodeiplus* counts the number of real nodes in the nodes
    resulting from applying the Xpath query *query* to *syntree*. If this count is
    greater or equal to *i*, the node is added to the results.

    It makes use of the function *getnodecount*, which is defined as follows:

    .. autofunction:: Sziplus::getnodecount
    '''
    clauses = syntree.xpath(query)
    results = []
    for clause in clauses:
        nodecount = getnodecount(clause)
        if nodecount >= i:
            results.append(clause)
    return results


def sziplus6(syntree: SynTree) -> List[SynTree]:
    '''
    The function *sziplus6* implements the TARSP language measure *6+*. It makes use of
    the function *sziplus*, which is applied to the *syntree* in combination with the
    integer *6*.

    .. autofunction:: Sziplus::sziplus

    * **Remark** The function *sziplus* was written in an early stage, but probably can now be better
    rewritten as a composed language measure.

    '''
    results = sziplus(syntree, 6)
    return results



import sys
from optparse import OptionParser

from sastadev.macros import expandmacros


def main():
    parser = OptionParser()
    parser.add_option("-f", "--file", dest="infilename",
                      help="File that contains the query to be expanded")
    parser.add_option("-q", "--query", dest="query",
                      help="query as a string")

    (options, args) = parser.parse_args()
    if options.infilename is not None:
        with open(options.infilename, 'r', encoding='utf8') as infile:
            query = infile.read()
    elif options.query is not None:
        query = options.query  # sys.argv[1]
    else:
        print('Specify a query (-q) or a filename containing a query (-f)', file=sys.stderr)
        exit(-1)
    expandedquery = expandmacros(query)
    outfilename = 'expandedqueries.txt'
    with open(outfilename, 'w', encoding='utf8') as of:
        print(expandedquery, file=of)


if __name__ == '__main__':
    main()



import os
from collections import defaultdict
from io import BytesIO
from shutil import copyfile, copyfileobj

from openpyxl import load_workbook

from sastadev.allresults import AllResults
from sastadev.conf import settings
from sastadev.forms import getformfilename

scoresheetname = 'STAP 1 - 5'
maxutt = 50
zerocount = 0
basexl = os.path.join(settings.SD_DIR, 'data', 'form_templates', 'STAP Excel VUmc 2018.xlsx')

NS = 'S001'
OS = 'S002'
PV = 'S003'
SGG = 'S004'
VT = 'S005'
VD = 'S006'
N = 'S007'
BvBep = 'S008'
zelfvnw3 = 'S009'
BBp = 'S010'
BBt = 'S011'
BBo = 'S012'

AG = 33
Ucol = 21
AF = 32

# order in the Excel sheet: NS	OS	PV	SGG	VT	VD	N	BvBep	zelf. vnw. 3	BB p	BB t	BB o
# i.e.
sorteditemlist = [NS, OS, PV, SGG, VT, VD, N, BvBep, zelfvnw3, BBp, BBt, BBo]


def data2rowtuples(data):
    # data is  a dictionary with key item and as value a counter with (uttid, count) items
    newdata = defaultdict(lambda: defaultdict(int))
    for item in data:
        for (uttid, count) in data[item].items():
            newdata[uttid][item] += count

    rowlist = []
    uttidlist = [uttid for uttid in newdata]
    sorteduttidlist = sorted(uttidlist)

    for uttid in sorteduttidlist:
        row = []
        for item in sorteditemlist:
            if item in newdata[uttid]:
                row.append(newdata[uttid][item])
            else:
                row.append(zerocount)
        rowlist.append((uttid, row))

    return rowlist


def makestapform(allresults, _, basexl=basexl, in_memory=False):
    if not in_memory:
        # copy the basexl to a new one with the appropriate name
        #(base, ext) = os.path.splitext(allresults.filename)
        #target = base + '_STAP-Form' + '.xlsx'
        target = getformfilename(allresults.filename, '_STAP-Form')

        copyfile(basexl, target)

        # open the workbook
        wb = load_workbook(filename=target)
    else:
        target = BytesIO()
        with open(basexl, 'rb') as source:
            copyfileobj(fsrc=source, fdst=target)
        wb = load_workbook(target)

    # gather the results

    # put the results in the right order
    rowlist = data2rowtuples(allresults.coreresults)

    ws = wb[scoresheetname]

    cols = ['U', 'V', 'W', 'X', 'Y', 'Z', 'AA', 'AB', 'AC', 'AD', 'AE', 'AF']
    # adapt the relevant sheet
    for (uttid, row) in rowlist:
        uttidrow = int(uttid) + 3
        xluttctr = ws.cell(column=AG, row=uttidrow).value
        uttidrowstr = str(uttidrow)
        if int(uttid) == xluttctr:
            for col, el in zip(cols, row):
                # special proviso for PV in column W
                if col == 'W':
                    el = el - 1
                cellkey = col + uttidrowstr
                ws[cellkey] = el
        else:
            settings.LOGGER.error('Unexpected utterance id encountered: {}'.format(uttid))

    # save the workbook
    wb.save(target)
    wb.close()

    # return the workbook- not needed
    return target


def test():
    coreresults = {NS: {'1': 3}, OS: {'1': 2, '2': 6}}
    postresults = {}
    allmatches = {}
    fn = 'STAP42.xml'
    analysedtrees = {}
    allresults = AllResults(0, coreresults, postresults, allmatches, fn, analysedtrees)
    fnbase, _ = os.path.splitext(fn)
    formxl = fnbase + '_form' + '.xlsx'
    makestapform(allresults, _)


if __name__ == '__main__':
    test()



from typing import Dict, List

from sastadev.query import pre_process
from sastadev.sastatypes import (AltCodeDict, ExactResult, ExactResultsDict,
                                 ExactResultsFilter, FileName,
                                 Item_Level2QIdDict, MethodName, QId, Query,
                                 QueryDict)

asta = 'asta'
stap = 'stap'
tarsp = 'tarsp'

validmethods = [asta, stap, tarsp]


def validmethod(rawmethod: str) -> bool:
    method = rawmethod.lower()
    result = method in validmethods
    return result


def allok(query: Query, xs: ExactResultsDict, x: ExactResult) -> bool:
    return True


class Method:
    '''

    The *Method* class has the following properties and methods:

* name : MethodName: name of the method
* queries : QueryDict = queries: a dictionary containing the queries (key is query id)
* item2idmap : Item_Level2QIdDict. Dictionary that maps an (item, level) tuple to the QueryId
* altcodes: AltCodeDict: dictionary with alternative codes and their mapping to the standard code
* postquerylist : List[QId]: list of query id’s for queries that are post or form queries
* methodfilename: FileName: filename of the file that contains the language measures
* defaultfilter: ExactResultsFilter: name of the function that acts as the default filter to regulate interaction between prequeries and core queries. By default it has the value allok


    '''

    def __init__(self, name: MethodName, queries: QueryDict, item2idmap: Item_Level2QIdDict,
                 altcodes: AltCodeDict, postquerylist: List[QId], methodfilename: FileName,
                 defaultfilter: ExactResultsFilter = allok):
        self.name: MethodName = name
        self.queries: QueryDict = queries
        self.defaultfilter: ExactResultsFilter = defaultfilter
        self.item2idmap: Item_Level2QIdDict = item2idmap
        self.altcodes: AltCodeDict = altcodes
        self.postquerylist: List[QId] = postquerylist
        self.methodfilename: FileName = methodfilename


def implies(a: bool, b: bool) -> bool:
    return (not a or b)


#filter specifies what passes the filter
def astadefaultfilter(query: Query, xrs: ExactResultsDict, xr: ExactResult) -> bool:
    return query.process == pre_process or \
        (implies('A029' in xrs, xr not in xrs['A029'])
         and implies('A045' in xrs, xr not in xrs['A045']))


defaultfilters: Dict[MethodName, ExactResultsFilter] = {}
defaultfilters[asta] = astadefaultfilter
defaultfilters[tarsp] = allok
defaultfilters[stap] = allok
